import os
import pynibs
import numpy as np
import unittest
import yaml
from inspect import getmembers, isfunction, getfullargspec

import pynibs


class TestRegressData(unittest.TestCase):
    """
    Test pynibs.regress_data()
    """
    np.random.seed(1)

    n_elm_small = 10
    # n_elm_large = 100000
    n_mep_small = 10
    # n_mep_large = 100
    n_signed = 25  # because function run_select_signed_data needs > 20 data points

    e_small = np.zeros((n_mep_small, n_elm_small))
    e_small[:] = 1
    e_small *= np.linspace(0.1, 5, n_mep_small)[:, np.newaxis]
    e_small *= np.linspace(0.1, 5, n_elm_small)

    e_signed = np.zeros((n_signed, n_signed))
    e_signed[:] = 1
    e_signed[0] = -1
    e_signed *= np.linspace(0.1, 5, n_signed)[:, np.newaxis]
    e_signed *= np.linspace(0.1, 5, n_signed)

    mep_small = np.linspace(0.1, 5, n_mep_small)

    mep_signed = np.linspace(0.1, 5, n_signed)

    con_small = np.array([[i, i + 1, i + 2] for i in range(n_elm_small)])

    def test_raise_assert_if_no_con(self):
        """
        When refit_discontinuities=True, con cannot be None
        """
        with self.assertRaises(AssertionError):
            pynibs.regress_data(self.e_small, self.mep_small)

    def test_all_default(self):
        r2 = pynibs.regress_data(self.e_small, self.mep_small, refit_discontinuities=False)
        assert r2.max() > .9
        assert r2.shape == (self.n_elm_small,)

    def test_return_fits(self):
        r2, fits = pynibs.regress_data(self.e_small, self.mep_small, refit_discontinuities=False, n_cpu=1,
                                       return_fits=True)
        assert r2.max() > .9
        assert r2.shape == (self.n_elm_small,)
        assert len(fits) == self.n_elm_small
        assert type(fits[0]) == dict
        for k in ['x0', 'r', 'amp', 'y0']:
            assert k in fits[0]

    def test_elm_idx_list(self):
        elm_idx_list = list(range(0, int(self.n_elm_small / 2)))
        r2 = pynibs.regress_data(self.e_small, self.mep_small, refit_discontinuities=False, n_cpu=1,
                                 elm_idx_list=elm_idx_list)
        assert r2.max() > .9
        assert r2.shape == (len(elm_idx_list),)

    def test_zap_idx(self):
        zap_idx = range(int(self.n_mep_small / 2), self.n_mep_small)
        r2 = pynibs.regress_data(self.e_small, self.mep_small, refit_discontinuities=False, n_cpu=1,
                                 zap_idx=zap_idx)
        assert r2.max() > .9
        assert r2.shape == (self.n_elm_small,)

    def test_element_list_linear(self):
        """
        Gives regress_data() an element_list AND fun=linear
        """
        element_list = [pynibs.Element(x=self.e_small[:, ele_id],
                                       y=self.mep_small,
                                       ele_id=ele_id,
                                       fun=pynibs.expio.fit_funs.linear,
                                       score_type="R2",
                                       select_signed_data=False,
                                       constants=None) for ele_id in range(self.n_elm_small)]
        r2, fits = pynibs.regress_data(self.e_small, self.mep_small, refit_discontinuities=False, n_cpu=1,
                                       element_list=element_list, return_fits=True)
        assert r2.max() > .9
        assert r2.shape == (self.n_elm_small,)
        # let's check if really linear fits have been used
        assert 'm' in fits[0]
        assert 'n' in fits[0]

    def test_refit_discontinuities(self):
        r2 = pynibs.regress_data(self.e_small, self.mep_small, refit_discontinuities=True, n_cpu=1,
                                 con=self.con_small)
        assert r2.max() > .9
        assert r2.shape == (self.n_elm_small,)

    def test_n_refit(self):
        r2 = pynibs.regress_data(self.e_small, self.mep_small, refit_discontinuities=False, n_cpu=1,
                                 n_refit=0)
        assert r2.max() > .9
        assert r2.shape == (self.n_elm_small,)

    def test_workhorses(self):  # and verbose
        # Suppress output message in run_select_signed_data()
        # suppress_text = io.StringIO()
        # sys.stdout = suppress_text

        r2 = pynibs.regress_data(self.e_small, self.mep_small, refit_discontinuities=False, n_cpu=4,
                                 n_refit=0, verbose=True)
        assert r2.max() > .9
        assert r2.shape == (self.n_elm_small,)

        # Enable output messages again
        # sys.stdout = sys.__stdout__

    def test_score_type(self):
        sr = pynibs.regress_data(self.e_small, self.mep_small, score_type='SR', refit_discontinuities=False, n_cpu=1,
                                 n_refit=0)
        assert sr.max() > .9
        assert sr.shape == (self.n_elm_small,)

        rho = pynibs.regress_data(self.e_small, self.mep_small, score_type='rho', refit_discontinuities=False, n_cpu=1,
                                  n_refit=0)
        assert rho.max() > .9
        assert rho.shape == (self.n_elm_small, 2)

        with self.assertRaises(NotImplementedError):
            pynibs.regress_data(self.e_small, self.mep_small, score_type='abc', refit_discontinuities=False, n_cpu=1,
                                n_refit=0)

    def test_select_signed_data(self):
        # Suppress output message in run_select_signed_data()
        # suppress_text = io.StringIO()
        # sys.stdout = suppress_text

        # Test with mostly positive data
        r2_pos = pynibs.regress_data(self.e_signed, self.mep_signed, select_signed_data=True,
                                     refit_discontinuities=False, n_refit=0)
        assert r2_pos.max() > 0.9
        assert r2_pos.shape == (self.n_signed,)

        # Test with mostly negative data
        r2_neg = pynibs.regress_data(-self.e_signed, self.mep_signed, select_signed_data=True,
                                     refit_discontinuities=False, n_refit=0)
        assert r2_neg.max() > 0.9
        assert r2_neg.shape == (self.n_signed,)

        # Enable output messages again
        # sys.stdout = sys.__stdout__

    # ***** Testing all regression functions: *****

    def test_dummy_fun(self):
        r2, fits = pynibs.regress_data(self.e_small, self.mep_small, refit_discontinuities=False, n_cpu=1,
                                       n_refit=0, fun=pynibs.expio.fit_funs.dummy_fun, return_fits=True)
        assert r2.shape == (self.n_elm_small,)
        assert 'a' in fits[0]

    def test_sigmoid(self):
        r2, fits = pynibs.regress_data(self.e_small, self.mep_small, refit_discontinuities=False, n_cpu=1,
                                       n_refit=0, fun=pynibs.expio.fit_funs.sigmoid, return_fits=True)
        assert r2.max() > 0.90
        assert r2.shape == (self.n_elm_small,)
        assert 'x0' in fits[0]
        assert 'r' in fits[0]
        assert 'amp' in fits[0]

    def test_sigmoid_negative_values(self):
        r2 = pynibs.regress_data(-self.e_small, self.mep_small, refit_discontinuities=False, n_cpu=1,
                                 n_refit=0, fun=pynibs.expio.fit_funs.sigmoid)
        assert r2.max() > 0.90
        assert r2.shape == (self.n_elm_small,)

    def test_sigmoid_log(self):
        r2, fits = pynibs.regress_data(self.e_small, self.mep_small, refit_discontinuities=False, n_cpu=1,
                                       n_refit=0, fun=pynibs.expio.fit_funs.sigmoid_log, return_fits=True)
        assert r2.max() > 0.9
        assert r2.shape == (self.n_elm_small,)
        assert 'x0' in fits[0]
        assert 'r' in fits[0]
        assert 'amp' in fits[0]

    def test_sigmoid4_log(self):
        r2, fits = pynibs.regress_data(self.e_small, self.mep_small, refit_discontinuities=False, n_cpu=1,
                                       n_refit=0, fun=pynibs.expio.fit_funs.sigmoid4_log, return_fits=True)
        assert r2.max() > 0.9
        assert r2.shape == (self.n_elm_small,)
        assert 'x0' in fits[0]
        assert 'r' in fits[0]
        assert 'amp' in fits[0]
        assert 'y0' in fits[0]

    def test_exp(self):
        r2, fits = pynibs.regress_data(self.e_small, self.mep_small, refit_discontinuities=False, n_cpu=1,
                                       n_refit=0, fun=pynibs.expio.fit_funs.exp0, return_fits=True)
        assert r2.max() > 0.90
        assert r2.shape == (self.n_elm_small,)
        assert 'x0' in fits[0]
        assert 'r' in fits[0]

    def test_raise_assert_if_false_fun(self):
        with self.assertRaises(NotImplementedError):
            pynibs.regress_data(self.e_small, self.mep_small, refit_discontinuities=False, n_cpu=1,
                                n_refit=0, fun=pynibs.expio.fit_funs.linear_log)

    # ***** Testing use of yaml-configuration files: *****

    def test_yaml_sigmoid4(self):
        yaml_config = os.path.join(pynibs.__datadir__, 'configuration_sigmoid4.yaml')
        with open(yaml_config, "r") as yamlfile:
            config = yaml.load(yamlfile, Loader=yaml.FullLoader)
        r2 = pynibs.regress_data(self.e_small, self.mep_small, refit_discontinuities=False, n_cpu=1,
                                 n_refit=0, fun=pynibs.expio.fit_funs.sigmoid4, **config)
        assert r2.max() > 0.90
        assert r2.shape == (self.n_elm_small,)

    def test_yaml_linear_MEP(self):
        yaml_config = os.path.join(pynibs.__datadir__, 'configuration_linear_MEP.yaml')
        with open(yaml_config, "r") as yamlfile:
            config = yaml.load(yamlfile, Loader=yaml.FullLoader)
        r2, fits = pynibs.regress_data(self.e_small, self.mep_small, refit_discontinuities=False, n_cpu=1,
                                       n_refit=0, fun=pynibs.expio.fit_funs.linear, return_fits=True, **config)
        assert r2.max() > 0.90
        assert r2.shape == (self.n_elm_small,)
        assert 'm' in fits[0]
        assert 'n' in fits[0]

    def test_yaml_linear_RT(self):
        yaml_config = os.path.join(pynibs.__datadir__, 'configuration_linear_RT.yaml')
        with open(yaml_config, "r") as yamlfile:
            config = yaml.load(yamlfile, Loader=yaml.FullLoader)
        r2, fits = pynibs.regress_data(self.e_small, self.mep_small, refit_discontinuities=False, n_cpu=1,
                                       n_refit=0, fun=pynibs.expio.fit_funs.linear, return_fits=True, **config)
        assert r2.max() > 0.90
        assert r2.shape == (self.n_elm_small,)
        assert 'm' in fits[0]
        assert 'n' in fits[0]

    def test_yaml_exp(self):
        yaml_config = os.path.join(pynibs.__datadir__, 'configuration_exp0.yaml')
        with open(yaml_config, "r") as yamlfile:
            config = yaml.load(yamlfile, Loader=yaml.FullLoader)
        r2, fits = pynibs.regress_data(self.e_small, self.mep_small, refit_discontinuities=False, n_cpu=1,
                                       n_refit=0, fun=pynibs.expio.fit_funs.exp0, return_fits=True, **config)
        assert r2.max() > 0.90
        assert r2.shape == (self.n_elm_small,)
        assert 'x0' in fits[0]
        assert 'r' in fits[0]


class TestSingleFit(unittest.TestCase):

    def setUp(self):
        self.x = x = np.array(range(1, 100))
        self.args = {'a': 1,
                     'x0': 0,
                     'r': .2,
                     'amp': 5,
                     'y0': 0,
                     'm': 5,
                     'n': 1}

    def test_single_fit(self):
        for fun_id, fun in getmembers(pynibs.expio.fit_funs):
            if isfunction(fun):
                if fun_id == 'dummy_fun':
                    continue

                fun_params = getfullargspec(fun)[0]
                print(f"Testing {fun_id} with: {fun_params}")
                fun_args = {k: self.args[k] for k in fun_params if k != 'x'}
                y = fun(self.x, **fun_args)

                fit = pynibs.single_fit(self.x, y, fun)
                for k in fun_params:
                    if k != 'x':
                        print(
                            f"\t{k}: org: {self.args[k]} | "
                            f"fit: {fit.params[k].value} | "
                            f"{np.isclose(self.args[k], fit.params[k].value, atol=.1)}")
                print(f"R2: {fit.rsquared}")


if __name__ == '__main__':
    unittest.main()



import numpy as np
import gdist
from matplotlib import pyplot as plt
from scipy.spatial.transform import Rotation as rot

import pynibs


def geodesic_dist(nodes, tris, source, source_is_node=True):
    """
    Returns geodesic distance in mm from all nodes to source node (or triangle).
    This is just a wrapper for the gdist package.

    Example
    -------

    .. code-block:: python

        with h5py.File(fn,'r') as f:
            tris = f['/mesh/elm/triangle_number_list'][:]
            nodes = f['/mesh/nodes/node_coord'][:]
        nodes_dist_ged, tris_dist_ged = geodesic_dist(nodes, tris, 3017)

        pynibs.write_data_hdf5(data_fn,
                            data=[tris_dist_ged, nodes_dist_ged],
                            data_names=["tris_dist_ged", "nodes_dist_ged"])
        pynibs.write_xdmf(data_fn,hdf5_geo_fn=fn)

    Parameters
    ----------
    nodes : np.ndarray
        (n_nodes,3) The nodes, determined by their x-y-z-coordinates.
    tris : np.ndarray
        (n_tris,3) Every triangle, determined by the node indices (vertices) that form it.
    source : np.ndarray(int) or int
        Geodesic distances of all nodes (or triangles) to this one will be computed.
    source_is_node : bool
        Whether source is a node idx or a triangle idx.

    Returns
    -------
    nodes_dist : np.ndarray
        (n_nodes,) Geodesic distance between source and all nodes in mm.
    tris_dist : np.ndarray
        (n_tris,) Geodesic distance between source and all triangles in mm.
    """
    if source_is_node:
        if type(source) is not np.ndarray:
            source = np.array([source])
        nodes_dist = gdist.compute_gdist(nodes.astype(np.float64),
                                         tris.astype(np.int32),
                                         source_indices=source.astype(np.int32))

    else:
        nodes = nodes.astype(np.float64)
        tris = tris.astype(np.int32)
        source = tris[source].astype(np.int32)
        nodes_dist = gdist.compute_gdist(nodes,
                                         tris,
                                         source_indices=source)
        # l = []
        # for n in source:
        #     l.append(np.mean(gdist.compute_gdist(nodes,
        #                                  tris,
        #                                  source_indices=np.array([n]).astype(np.int32))[tris],axis=1))
        # # a = np.array(l)
        # nodes_dist = np.mean(l,axis=0)
        # a.shape
    tris_dist = np.mean(nodes_dist[tris], axis=1)

    return nodes_dist, tris_dist


def euclidean_dist(nodes, tris, source, source_is_node=True):
    """
    Returns euclidean distance of all nodes to source node (triangle).
    This is just a wrapper for the gdist package.

    Example
    -------
    .. code-block:: python

        with h5py.File(fn,'r') as f:
            tris = f['/mesh/elm/triangle_number_list'][:]
            nodes = f['/mesh/nodes/node_coord'][:]
        nodes_dist_euc, tris_dist_euc = euclidean_dist(nodes, tris, 3017)

        pynibs.write_data_hdf5(data_fn,
                            data=[tris_dist_euc, nodes_dist_euc],
                            data_names=["tris_dist_euc", "nodes_dist_euc", "])
        pynibs.write_xdmf(data_fn,hdf5_geo_fn=fn)

    Parameters
    ----------
    nodes : np.ndarray
        (n_nodes,3) The nodes, determined by their x-y-z-coordinates.
    tris : np.ndarray
        (n_tris,3) Every triangle, determined by the node indices (vertices) that form it.
    source : np.ndarray(int) or int
        Euclidean distances of all nodes (or triangles) to this one will be computed.
    source_is_node : bool
        Whether source is a node idx or a triangle idx.

    Returns
    -------
    nodes_dist : np.ndarray
        (n_nodes,) Geodesic distance between source and all nodes in mm.
    tris_dist : np.ndarray
        (n_tris,) Geodesic distance between source and all triangles in mm.
    """
    if source_is_node:
        nodes_dist = np.linalg.norm(nodes - nodes[source], axis=1)
    else:
        nodes_dist = np.zeros(nodes.shape[0], )
        for node in nodes[tris[source]]:
            nodes_dist += np.linalg.norm(nodes - node, axis=1)
        nodes_dist /= 3
    tris_dist = np.mean(nodes_dist[tris], axis=1)

    return nodes_dist, tris_dist


def nrmsd(array, array_ref, error_norm="relative", x_axis=False):
    """
    Determine the normalized root-mean-square deviation between input data and reference data.

    Notes
    -----
    nrmsd = np.sqrt(1.0 / n_points * np.sum((data - data_ref) ** 2)) / (max(array_ref) - min(array_ref))

    Parameters
    ----------
    array : np.ndarray
        input data [ (x), y0, y1, y2 ... ].
    array_ref : np.ndarray
        reference data [ (x_ref), y0_ref, y1_ref, y2_ref ... ].
        if array_ref is 1D, all sizes have to match.
    error_norm : str, optional, default="relative"
        Decide if error is determined "relative" or "absolute".
    x_axis : bool, default: False
        If True, the first column of array and array_ref is interpreted as the x-axis, where the data points are
        evaluated. If False, the data points are assumed to be at the same location.

    Returns
    -------
    normalized_rms : np.ndarray of float
        ([array.shape[1]]) Normalized root-mean-square deviation between the columns of array and array_ref.
    """

    n_points = array.shape[0]

    if x_axis:
        # handle different array lengths
        if len(array_ref.shape) == 1:
            array_ref = array_ref[:, None]
        if len(array.shape) == 1:
            array = array[:, None]

        # determine number of input arrays
        if array_ref.shape[1] == 2:
            n_data = array.shape[1] - 1
        else:
            n_data = array.shape[1]

        # interpolate array on array_ref data if necessary
        if array_ref.shape[1] == 1:
            data = array
            data_ref = array_ref
        else:
            # crop reference if it is longer than the axis of the data
            data_ref = array_ref[(array_ref[:, 0] >= min(array[:, 0])) & (array_ref[:, 0] <= max(array[:, 0])), 1]
            array_ref = array_ref[(array_ref[:, 0] >= min(array[:, 0])) & (array_ref[:, 0] <= max(array[:, 0])), 0]

            data = np.zeros([len(array_ref), n_data])
            for i_data in range(n_data):
                data[:, i_data] = np.interp(array_ref, array[:, 0], array[:, i_data + 1])
    else:
        data_ref = array_ref
        data = array

    # determine "absolute" or "relative" error
    if error_norm == "relative":
        # max_min_idx = np.isclose(np.max(data_ref, axis=0), np.min(data_ref, axis=0))
        delta = np.max(data_ref, axis=0) - np.min(data_ref, axis=0)

        # if max_min_idx.any():
        #     delta[max_min_idx] = max(data_ref[max_min_idx])
    elif error_norm == 'absolute':
        delta = 1
    else:
        raise NotImplementedError

    # determine normalized rms deviation and return
    normalized_rms = np.sqrt(1.0 / n_points * np.sum((data - data_ref) ** 2, axis=0)) / delta

    return normalized_rms


def nrmse(array, array_ref, x_axis=False):
    """
    Determine the normalized root-mean-square deviation between input data and reference data.

    nrmse = np.linalg.norm(array - array_ref) / np.linalg.norm(array_ref)

    Parameters
    ----------
    array : np.ndarray
        input data [ (x), y0, y1, y2 ... ].
    array_ref : np.ndarray
        reference data [ (x_ref), y0_ref, y1_ref, y2_ref ... ].
        if array_ref is 1D, all sizes have to match.
    error_norm : str, optional, default="relative"
        Decide if error is determined "relative" or "absolute".
    x_axis : bool, default: False
        If True, the first column of array and array_ref is interpreted as the x-axis, where the data points are
        evaluated. If False, the data points are assumed to be at the same location.

    Returns
    -------
    nrmse: np.ndarray of float
        ([array.shape[1]]) Normalized root-mean-square deviation between the columns of array and array_ref.
    """

    n_points = array.shape[0]

    if x_axis:
        # handle different array lengths
        if len(array_ref.shape) == 1:
            array_ref = array_ref[:, None]
        if len(array.shape) == 1:
            array = array[:, None]

        # determine number of input arrays
        if array_ref.shape[1] == 2:
            n_data = array.shape[1] - 1
        else:
            n_data = array.shape[1]

        # interpolate array on array_ref data if necessary
        if array_ref.shape[1] == 1:
            data = array
            data_ref = array_ref
        else:
            # crop reference if it is longer than the axis of the data
            data_ref = array_ref[(array_ref[:, 0] >= min(array[:, 0])) & (array_ref[:, 0] <= max(array[:, 0])), 1]
            array_ref = array_ref[(array_ref[:, 0] >= min(array[:, 0])) & (array_ref[:, 0] <= max(array[:, 0])), 0]

            data = np.zeros([len(array_ref), n_data])
            for i_data in range(n_data):
                data[:, i_data] = np.interp(array_ref, array[:, 0], array[:, i_data + 1])
    else:
        data_ref = array_ref
        data = array

    # determine normalized rms deviation and return
    nrmse = np.linalg.norm(data - data_ref, axis=0) / np.linalg.norm(data_ref, axis=0)

    return nrmse


def c_map_comparison(c1, c2, t1, t2, nodes, tris):
    """
    Compares two c-maps in terms of NRMSD and calculates the geodesic distance between the hotspots.

    Parameters
    ----------
    c1 : np.ndarray of float
        (n_ele) First c-map.
    c2 : np.ndarray of float
        (n_ele) Second c-map (reference).
    t1 : np.ndarray of float
        (3) Coordinates of the hotspot in the first c-map.
    t2 : np.ndarray of float
        Coordinates of the hotspot in the second c-map.
    nodes : np.ndarray of float
        (n_nodes, 3) Node coordinates
    tris : np.ndarray of float
        (n_tris, 3) Connectivity of ROI elements

    Returns
    -------
    nrmsd : float
        Normalized root-mean-square deviation between the two c-maps in (%).
    gdist : float
        Geodesic distance between the two hotspots in (mm).
    """
    # determine NRMSD between two c-maps
    nrmsd_ = nrmsd(array=c1, array_ref=c2, error_norm="relative", x_axis=False) * 100

    # determine geodesic distance between hotspots
    tris_center = np.mean(nodes[tris,], axis=1)

    t1_idx = np.argmin(np.linalg.norm(tris_center - t1, axis=1))
    t2_idx = np.argmin(np.linalg.norm(tris_center - t2, axis=1))
    gdists = geodesic_dist(nodes=nodes, tris=tris, source=t2_idx, source_is_node=False)[1]
    gdist_ = gdists[t1_idx]

    return nrmsd_, gdist_


def calc_tms_motion_params(coil_positions, reference=None):
    """
    Computes motion parameters for a set of ``coil_positions``, i.e., coil shifts in [mm] in RAS coordinate system and
    rotations (pitch, roll, yaw) in [°].
    Motion is computed w.r.t. the first ('absolute') and to the previous ('relative') stimulation.

    Position shifts are quantified with respect to the subject/nifti-specific RAS coordinate system.
    Rorational changes are quantified with respect to the coil axes as follows:

        * pitch: rotation around left/right axis of coil
        * roll: rotation around coil handle axis
        * yaw: rotation around axis from center of coil towards head
    Motion parameters for first coil position are set to 0.

    Parameters
    ----------
    coil_positions : np.ndarray of float
        (4, 4, n_pulses).
    reference : np.ndarray of float, optional
        (4, 4) Reference coil placement. If None (default), the first placement from ``coil_positions`` is used.

    Returns
    -------
    pos_diff_abs : np.ndarray
        (3, n_pulses) Absolute position differences (R, A, S).
    euler_rots_abs : np.ndarray
        (3, n_pulses) Absolute rotation angles in euler angles (alpha, beta, gamma).
    pos_diff_rel : np.ndarray
        (3, n_pulses) Relative position differences (R, A, S).
    euler_rots_rel : np.ndarray
        (3, n_pulses) Relative rotation angles in euler angles (alpha, beta, gamma).
    """
    np.set_printoptions(suppress=True)
    if reference is None:
        i_ref = 0
        while np.isnan(coil_positions[:, :, i_ref]).any():
            i_ref += 1
        reference_pos = coil_positions[0:3, 3, i_ref].T
        reference_rot = coil_positions[0:3, 0:3, i_ref]

        # first rotation is zero because it is the reference
        euler_rots_abs = [[0, 0, 0]]
    else:
        reference = np.squeeze(reference)
        reference_pos = reference[0:3, 3].T
        reference_rot = reference[0:3, 0:3]

        # compute first rotation towards reference
        rotmat_abs = pynibs.bases2rotmat(reference_rot, coil_positions[0:3, 0:3, 0])
        if np.isnan(rotmat_abs).any():
            euler_rots_abs = [[np.nan, np.nan, np.nan]]
        else:
            euler_rots_abs = [rot.from_matrix(rotmat_abs).as_euler('xyz', degrees=True).tolist()]

    # compute absolute and relative position differences
    pos_diff_abs = (reference_pos - coil_positions[0:3, 3, :].T).T
    pos_diff_rel = np.diff(coil_positions[0:3, 3, :], 1, axis=1, prepend=coil_positions[0:3, 3, 0, np.newaxis])

    # bring position differences into reference coordinate system
    pos_diff_abs_rot = reference_rot.T @ pos_diff_abs
    pos_diff_rel_rot = reference_rot.T @ pos_diff_rel

    n_coilpos = coil_positions.shape[2]
    euler_rots_rel = [[0, 0, 0]]
    for i in range(0, n_coilpos - 1):
        rotmat_abs = pynibs.bases2rotmat(reference_rot, coil_positions[0:3, 0:3, i + 1])
        if np.isnan(rotmat_abs).any():
            euler_abs = [np.nan, np.nan, np.nan]
        else:
            euler_abs = rot.from_matrix(rotmat_abs).as_euler('xyz', degrees=True).tolist()
        euler_rots_abs.append(euler_abs)

        rotmat_rel = pynibs.bases2rotmat(coil_positions[0:3, 0:3, i], coil_positions[0:3, 0:3, i + 1])
        if np.isnan(rotmat_rel).any():
            euler_rel = [np.nan, np.nan, np.nan]
        else:
            euler_rel = rot.from_matrix(rotmat_rel).as_euler('xyz', degrees=True).tolist()
        euler_rots_rel.append(euler_rel)

    euler_rots_rel = np.array(euler_rots_rel)
    euler_rots_abs = np.array(euler_rots_abs)

    return pos_diff_abs_rot, euler_rots_abs.T, pos_diff_rel_rot, euler_rots_rel.T


def plot_tms_motion_parameter(pos_diff, euler_rots, pcd=None, fname=None):
    """
    Plots TMS coil motion parameters.

    .. figure:: ../../doc/images/pcd.png
       :scale: 80 %
       :alt: TMS motion parameters and PCD

       Pulsewise Coil Displacement (PCD) is a compound metric to quantify TMS coil movements. Data for a double (space)
       cTBS400 protocol is shown (with a break after 200 bursts).

    Parameters
    ----------
    pos_diff : np.ndarray
        (3, n_pulses) Position differences (R, A, S).
    euler_rots : np.ndarray
        (3, n_pulses) Rotation angles in euler angles (alpha, beta, gamma).
    pcd : np.ndarray, optional
        (n_pulses,) pulsewise coil displacements.
    fname : str, optional
        Filename to save plot.

    Returns
    -------
    axes : matplotlib.pyplot.axes
        Figure axes.
    """
    if pcd is not None:
        n_plot_rows = 3
    else:
        n_plot_rows = 2
    n_coilpos = pos_diff.shape[1]
    fig, axes = plt.subplots(n_plot_rows, 2)

    fig.suptitle('TMS coil displacement')

    # let's order the plotting based on the data to have the dimension with the fewest changes on top
    order_pos = np.argsort(np.std(pos_diff, axis=1))[::-1]
    order_rot = np.argsort(np.std(euler_rots, axis=1))[::-1]
    colors_pos = ['#12263A', '#50858B', '#99EDCC']
    colors_rot = ['#725752', '#F2A541', '#F8DF8C']
    labels_pos = ['X', 'Y', 'Z']
    labels_rot = ['yaw', 'pitch', 'roll']

    for i in order_pos:
        axes[0, 0].plot(range(0, n_coilpos), pos_diff[i, :], c=colors_pos[i], label=labels_pos[i], linewidth=1)
        axes[0, 1].plot(range(0, n_coilpos), np.nancumsum(pos_diff[i, :]), c=colors_pos[i], label=labels_pos[i],
                        linewidth=1)
    for i in order_rot:
        axes[1, 0].plot(range(0, n_coilpos), euler_rots[i, :], c=colors_rot[i], label=labels_rot[i], linewidth=1)
        axes[1, 1].plot(range(0, n_coilpos), np.nancumsum(euler_rots[i, :]), c=colors_rot[i], label=labels_rot[i],
                        linewidth=1)
    np.set_printoptions(suppress=True)
    axes[0, 0].set_title('Relative movement [mm]')
    axes[1, 0].set_title('Relative rotation [°]')
    axes[0, 1].set_title('Cumulative movement [mm]')
    axes[1, 1].set_title('Cumulative rotation [°]')
    fig.delaxes(axes[2, 1])
    if pcd is not None:
        axes[2, 0].plot(range(0, n_coilpos), pcd, c='#843E84', label='PCD',
                        linewidth=1)
        axes[2, 0].set_title('Pulsewise coil displacement [AU]')

        # add some stats to plot
        plt.figtext(0.505, 0.09, f"       PCD\n"
                                 f"   Max: {np.nanmax(pcd).round(2):6.2f}\n"
                                 f"  Mean: {np.nanmean(pcd).round(2):6.2f}\n"
                                 f"Median: {np.nanmedian(pcd).round(2):6.2f}\n"
                                 f"    SD: {np.nanstd(pcd).round(2):6.2f}\n"
                                 f"n stim: {(~np.isnan(pcd)).sum(): >6}\n"
                                 f"untracked: {(np.isnan(pcd)).sum(): >3}",
                    family='monospace'
                    )

    fig.tight_layout()
    handles_tmp, labels_tmp = axes[0, 0].get_legend_handles_labels()
    fig.legend(handles_tmp, labels_tmp, ncols=1, loc=[0.7, 0.08], title='Shift')

    handles_tmp, labels_tmp = axes[1, 0].get_legend_handles_labels()
    fig.legend(handles_tmp, labels_tmp, ncols=1, loc=[0.83, 0.08], title='Rotation')

    # handles_tmp, labels_tmp = axes[2, 0].get_legend_handles_labels()
    # fig.legend(handles_tmp, labels_tmp, ncols=1, loc=[0.85, 0.12], title='PCD')

    if fname is not None:
        plt.savefig(fname)

    return axes


def compute_pcd(delta_pos, delta_rot, skin_cortex_distance=20):
    """
    Computes _P_ulsewise _C_oil _D_isplacements (PCD) based on 3 position parameters (``delta_pos``) and
    3 rotation parmeters (``delta_rot``).

    The coil rotations (in euler angles) are transformed into a positional change projected on the cortex based on
    ``skin_cortex_distance`` as a proxy for the (local) change of the stimulation.
    ``delta_pos`` and ``delta_rot`` are expected to quantify motion w.r.t. the target coil position/roation, for example
    the absolute deltas to the first stimulation.

    Axes definitions
    ----------------
    ``delta_pos`` and ``delta_rot`` are supposed to follow this axes definition (SimNIBS):

    - delta_pos
        - X and Y: coil movement tangential to head surface
        - Z: coil movement perpendicular to head surface

    - delta_rot
        * pitch: rotation around left/right axis of coil
        * roll: rotation around coil handle axis
        * yaw: rotation around axis from center of coil towards head

    .. figure:: ../../doc/images/coil_axes_definition.png
       :scale: 80 %
       :alt: Coil axes definition, following SimNIBS conventions.


    Example
    -------
    .. code-block:: python

        # get TriggerMarkers from a Localite session
        mats = pynibs.read_triggermarker_localite(tm_fn)[0]

        # calculate absolute and relative coil displacements
        delta_pos_abs, delta_rot_abs, delta_pos_rel, delta_rot_rel = pynibs.calc_tms_motion_params(mats)

        # compute PCD
        pcd, delta_pos, delta_rot = pynibs.compute_pcd(delta_pos_abs, delta_rot_abs)

        # plot movement
        axes = pynibs.plot_tms_motion_parameter(pos_diff_rel, euler_rots_rel, pcd)
        matplotlib.pyplot.show()

    Parameters
    ----------
    delta_pos : np.ndarray
        (3, n_pulses) Absolute position differences (R, A, S).
    delta_rot : np.ndarray
        (3, n_pulses) Absolute rotation angles in euler angles (alpha, beta, gamma).
    skin_cortex_distance : int, default: 20
        Cortical depth of target to adjust coil rotations for.

    Returns
    -------
    pcd : np.ndarray
        Pulsewise coil displacements.
    delta_pos : np.ndarray
        (n_pulses, ) sum(abs(delta_pos)) per pulse.
    delta_rot : np.ndarray
        (n_pulses, ) sum(abs(delta_rot projected by skin_cortex_distance)) per pulse.
    """
    # delta_pos_summed = np.sum(np.abs(delta_pos[:2, :]), axis=0)
    delta_pos_summed = np.sqrt(np.sum(delta_pos[: 2, :] ** 2, axis=0))
    delta_pos_summed = np.squeeze(delta_pos_summed + delta_pos[2, :] ** 2 * np.sign(delta_pos[2, :]))

    plane_n = np.array([0, 0, 1])
    plane_p = np.array([0, 0, skin_cortex_distance])  # Any point on the plane

    # compute shift of the center of coil ray
    ray_origin = np.array([0, 0, 0])  # Any point along the ray
    delta_rot_a = []
    for idx, rots in enumerate(delta_rot.T):
        # apply x and y rotation to [0,0,1]
        # rots[2] = 0
        r = rot.from_rotvec(rots, degrees=True)
        ray_dir = r.apply([0, 0, 1])
        intersec = pynibs.intersection_vec_plan(ray_dir, ray_origin, plane_n, plane_p, eps=1e-6)

        # compute distance from new intersection with [0, 0, scd]
        delta_rot_a.append(np.linalg.norm(plane_p - intersec))

    # ray shift based on orientation doesn't include any rotations around z axis.
    z_rotation_deltas = np.abs((skin_cortex_distance) * np.sin(np.deg2rad(delta_rot[2, :])))
    delta_rot = np.array(delta_rot_a) + z_rotation_deltas
    return delta_pos_summed + delta_rot, delta_pos_summed, delta_rot



import pynibs
import unittest
from scipy.spatial.transform import Rotation as rot
import numpy as np


class TestIntersectionVecPlan(unittest.TestCase):
    def test_intersection(self):
        r = rot.from_rotvec([45, 0, 0], degrees=True)

        # Define plane
        plane_n = np.array([0, 0, 1])
        plane_p = np.array([0, 0, 20])  # Any point on the plane

        # Define ray
        ray_dir = r.apply([0, 0, 1])  # np.array([0, -1, -1])
        ray_origin = np.array([0, 0, 0])  # Any point along the ray

        ret = pynibs.intersection_vec_plan(ray_dir, ray_origin, plane_n, plane_p, eps=1e-6)
        assert np.all(np.isclose(np.array([0., -20., 20.]), np.array(ret)))


if __name__ == '__main__':
    unittest.main()



import os
import pynibs
import unittest
import numpy as np


class TestFindNearest(unittest.TestCase):
    monot_arr = np.array([0, 1, 2, 3, 5, 10])
    nonmonot_arr = np.array([10, 0, 1, 2, 3, 5, 10])
    value = 5

    def test_01_find_nearest(self):
        idx = pynibs.find_nearest(array=self.monot_arr, value=self.value)
        assert idx == 4

    def test_02_find_nearest_fail(self):
        idx = pynibs.find_nearest(array=self.nonmonot_arr, value=self.value)
        assert idx == -1


class TestCalcSampleSphere(unittest.TestCase):
    mesh_hdf5_fn = os.path.join(pynibs.__testdatadir__, 'geo.hdf5')
    mesh = pynibs.load_mesh_hdf5(mesh_hdf5_fn)

    def test_01_sample_sphere_even(self):
        with self.assertRaises(AssertionError):
            pynibs.sample_sphere(n_points=10, r=10)

    def test_02_sample_sphere_odd(self):
        points = pynibs.sample_sphere(n_points=11, r=10)
        assert points.shape == (11, 3)


class TestCalcTriSurface(unittest.TestCase):
    mesh_hdf5_fn = os.path.join(pynibs.__testdatadir__, 'geo.hdf5')
    mesh = pynibs.load_mesh_hdf5(mesh_hdf5_fn)

    def test_01_calc_tri_volume(self):
        tri_surf = pynibs.calc_tri_surface(points=self.mesh.points[self.mesh.triangles])
        assert tri_surf.size == self.mesh.triangles.shape[0]
        assert (tri_surf > 0).all()


class TestCalcTetVolume(unittest.TestCase):
    mesh_hdf5_fn = os.path.join(pynibs.__testdatadir__, 'geo.hdf5')
    mesh = pynibs.load_mesh_hdf5(mesh_hdf5_fn)

    def test_01_calc_tet_volume_abs(self):
        tet_vol = pynibs.calc_tet_volume(points=self.mesh.points[self.mesh.tetrahedra])
        assert tet_vol.size == self.mesh.tetrahedra.shape[0]
        assert (tet_vol > 0).all()

    def test_02_calc_tet_volume_no_abs(self):
        tet_vol = pynibs.calc_tet_volume(points=self.mesh.points[self.mesh.tetrahedra],
                                         abs=False)
        assert tet_vol.size == self.mesh.tetrahedra.shape[0]
        assert (tet_vol < 0).any()


class TestGetSphere(unittest.TestCase):
    """
    Test pynibs.mesh.utils.* functions.
    """
    mesh_hdf5_fn = os.path.join(pynibs.__testdatadir__, 'geo.hdf5')
    target_empty = [10, 20, 30]
    target_full_tris = [0., 45.2227265, 71.97155693]
    target_full_tets = [9.51698047, 84.46553784, 0.]
    radius = 3

    def test_01_get_sphere_fails(self):
        # no mesh and no mesh_fn
        with self.assertRaises(AssertionError):
            pynibs.get_sphere(mesh=None,
                              mesh_fn=None,
                              target=self.target_empty,
                              radius=self.radius)

        # no target
        with self.assertRaises(AssertionError):
            pynibs.get_sphere(mesh=None, mesh_fn=self.mesh_hdf5_fn)

        # wrong element type
        with self.assertRaises(ValueError):
            # no radius, return nearest element
            pynibs.get_sphere(mesh=None,
                              mesh_fn=self.mesh_hdf5_fn,
                              target=self.target_empty,
                              elmtype='square')

    def test_02_get_sphere_nearest_tri(self):
        # no radius, return nearest element
        nearest_elm = pynibs.get_sphere(mesh=None,
                                        mesh_fn=self.mesh_hdf5_fn,
                                        target=self.target_empty)
        assert nearest_elm == 681

        nearest_elm = pynibs.get_sphere(mesh=None,
                                        mesh_fn=self.mesh_hdf5_fn,
                                        target=self.target_empty,
                                        elmtype='Triangle')
        assert nearest_elm == 681

    def test_02_get_sphere_nearest_tet(self):
        # no radius, return nearest element
        nearest_elm = pynibs.get_sphere(mesh=None,
                                        mesh_fn=self.mesh_hdf5_fn,
                                        target=self.target_empty,
                                        elmtype='tets')
        assert nearest_elm == 20232

        nearest_elm = pynibs.get_sphere(mesh=None,
                                        mesh_fn=self.mesh_hdf5_fn,
                                        target=self.target_empty,
                                        elmtype='Tetrahedron')
        assert nearest_elm == 20232

        # return sphere, but should be empty
        sphere_empty = pynibs.get_sphere(mesh=None,
                                         mesh_fn=self.mesh_hdf5_fn,
                                         target=self.target_empty,
                                         radius=5)
        assert sphere_empty.size == 0

    def test_02_get_sphere_tris(self):
        # return tri sphere, should not be empty
        sphere_full_tris = pynibs.get_sphere(mesh=None,
                                             mesh_fn=self.mesh_hdf5_fn,
                                             target=self.target_full_tris,
                                             radius=10)
        assert sphere_full_tris.size == 13

    def test_03_get_sphere_tets(self):
        # return tet sphere, should not be empty
        sphere_full_tets = pynibs.get_sphere(mesh=None,
                                             mesh_fn=self.mesh_hdf5_fn,
                                             target=self.target_full_tets,
                                             radius=10,
                                             elmtype='tets')
        assert sphere_full_tets.size == 50


if __name__ == '__main__':
    unittest.main()



import os
import h5py
import pickle
import warnings
import numpy as np
import pynibs


class Subject:
    """
    Subject containing subject specific information, like mesh, roi, uncertainties, plot settings.

    Attributes
    ----------
    self.id : str
        Subject id.

    Notes
    -----
    **Initialization**

    .. code-block:: python

        sub = pynibs.subject(subject_ID, mesh)

    **Parameters**

    id : str
        Subject id.
    fn_mesh : str
         .msh or .hdf5 file containing the mesh information.

    **Subject.seg, segmentation information dictionary**

    fn_lh_wm : str
        Filename of left hemisphere white matter surface.
    fn_rh_wm : str
        Filename of right hemisphere white matter surface.
    fn_lh_gm : str
        Filename of left hemisphere grey matter surface.
    fn_rh_gm : str
        Filename of right hemisphere grey matter surface.
    fn_lh_curv : str
        Filename of left hemisphere curvature data on grey matter surface.
    fn_rh_curv : str
        Filename of right hemisphere curvature data on grey matter surface.

    **Subject.mri, mri information dictionary**

    fn_mri_T1 : str
        Filename of T1 image.
    fn_mri_T2 : str
        Filename of T2 image.
    fn_mri_DTI : str
        Filename of DTI dataset.
    fn_mri_DTI_bvec : str
        Filename of DTI bvec file.
    fn_mri_DTI_bval : str
        Filename of DTI bval file.
    fn_mri_conform : str
        Filename of conform T1 image resulting from SimNIBS mri2mesh function.

    **Subject.ps, plot settings dictionary**

    see plot functions in para.py for more details

    **Subject.exp, experiment dictionary**

    info : str
        General information about the experiment.
    date : str
        Date of experiment (e.g. 01/01/2018).
    fn_tms_nav : str
        Path to TMS navigator folder.
    fn_data : str
        Path to data folder or files.
    fn_exp_csv : str
        Filename of experimental data .csv file containing the merged experimental data information.
    fn_coil : str
        Filename of .ccd or .nii file of coil used in the experiment (contains ID).
    fn_mri_nii : str
        Filename of MRI .nii file used during the experiment.
    cond : str or list of str
        Conditions in the experiment in the recorded order (e.g. ['PA-45', 'PP-00']).
    experimenter : str
        Name of experimenter who conducted the experiment.
    incidents : str
        Description of special events occurred during the experiment.

    **Subject.mesh, mesh dictionary**

    info : str
        Information about the mesh (e.g. dicretization, etc.).
    fn_mesh_msh : str
        Filename of the .msh file containing the FEM mesh.
    fn_mesh_hdf5 : str
        Filename of the .hdf5 file containing the FEM mesh.
    seg_idx : int
        Index indicating to which segmentation dictionary the mesh belongs.

    **Subject.roi region of interest dictionary**

    type : str
        Specify type of ROI ('surface', 'volume')
    info : str
        Info about the region of interest, e.g. "M1 midlayer from freesurfer mask xyz".
    region : list of str or float
        Filename for freesurfer mask or ``[[X_min, X_max], [Y_min, Y_max], [Z_min, Z_max]]``.
    delta : float
        Distance parameter between WM and GM (0 -> WM, 1 -> GM) (for surfaces only).
    """
    def __init__(self, subject_id, subject_folder):
        self.id = subject_id  # subject id
        self.subject_folder = subject_folder  # folder containing subject information
        self.mri = []  # list containing mri information
        self.mesh = {}  # dict containing mesh information
        self.exp = {}  # dict containing information about conducted experiments
        self.ps = []  # list containing plot settings
        self.roi = {}  # dict containing the roi information

    def __str__(self):
        """ Overload method to allow print(subject_object)"""
        ret = f'{"=" * 64}\n' \
              f'Subject ID : {self.id}\n' \
              f'Folder     : {self.subject_folder}\n' \
              f'Meshes     : {len(self.mesh.items())}\n' \
              f'Experiments: {len(self.exp.items())}\n'
        ret += f'{"=" * 64}\n\n'
        ret += '------------------------------MRI------------------------------\n'
        if not self.mri:
            ret += 'None\n'
        else:
            for idx_mesh, mri in enumerate(self.mri):
                ret += f"|- MRI[{idx_mesh}]\n"
                for k, d in mri.items():
                    if k == list(mri.keys())[-1]:
                        pref = '└-'
                    else:
                        pref = '|-'
                    ret += f'   {pref}  {k: >19}: {d}\n'

        ret += '\n------------------------------MESH------------------------------\n'
        if not self.mesh:
            ret += 'None\n'
        else:
            for idx_mesh, mesh in self.mesh.items():
                ret += f"|- Mesh name: {idx_mesh}\n"

                # plot roi keys
                if idx_mesh in self.roi.keys():
                    for r_idx, roi in self.roi[idx_mesh].items():
                        key_list_roi = []
                        ret += f'  |-- ROI: {r_idx}\n'
                        for k, d in roi.items():
                            if d is not None and k != 'name' and k != 'mesh_name':
                                key_list_roi.append(k)
                        for k in key_list_roi:
                            if k == key_list_roi[-1]:
                                pref = '└-'
                            else:
                                pref = '|-'
                            ret += f'    {pref} {k: >19}: {roi[k]}\n'

                # plot mesh keys
                key_list_mesh = []
                for k, d in mesh.items():
                    if d is not None and k != 'name' and k != 'subject_id':
                        key_list_mesh.append(k)
                for k in key_list_mesh:
                    if k == key_list_mesh[-1]:
                        pref = '└-'
                    else:
                        pref = '|-'
                    ret += f'  {pref} {k: >19}: {mesh[k]}\n'

                ret += "\n"
        ret += '\n----------------------------Experiments-------------------------\n'
        if not self.exp:
            ret += 'None\n'
        else:
            for idx_exp, exp in self.exp.items():
                ret += f"|- Exp name: {idx_exp}\n"
                key_list_exp = []
                for k, d in exp.items():
                    if d is not None and k != 'name' and k != 'subject_id':
                        key_list_exp.append(k)
                for k in key_list_exp:
                    if k == key_list_exp[-1]:
                        pref = '└-'
                    else:
                        pref = '|-'
                    ret += f'  {pref} {k: >19}: {exp[k]}\n'
        return ret

    def add_mesh_info(self, mesh_dict):
        """
        Adding filename information of the mesh to the subject object (multiple filenames possible).

        Parameters
        ----------
        mesh_dict : dict or list of dict
            Dictionary containing the mesh information.

        Notes
        -----
        **Adds Attributes**

        Subject.mesh : list of dict
            Dictionaries containing the mesh information.
        """
        if type(mesh_dict) is list:
            for mesh in mesh_dict:
                self.mesh.append(mesh)

        else:
            self.mesh = mesh_dict

    def add_roi_info(self, roi_dict):
        """
        Adding ROI (surface) information of the mesh with mesh_index to the subject object (multiple ROIs possible).

        Parameters
        ----------
        roi_dict : dict of dict or list of dict
            Dictionary containing the ROI information of the mesh with mesh_index ``[mesh_idx][roi_idx]``.

        Notes
        -----
        **Adds Attributes**

        Subject.mesh[mesh_index].roi : list of dict
            Dictionaries containing ROI information.
        """
        for mesh_idx in roi_dict.keys():
            self.roi[mesh_idx] = dict()

            for roi_idx in roi_dict[mesh_idx].keys():
                self.roi[mesh_idx][roi_idx] = roi_dict[mesh_idx][roi_idx]

    def add_plotsettings(self, ps_dict):
        """
        Adding ROI information to the subject object (multiple ROIs possible).

        Parameters
        ----------
        ps_dict : dict or list of dict
            Dictionary containing plot settings of the subject.

        Notes
        -----
        **Adds Attributes**

        Subject.ps : list of dict
            Dictionary containing plot settings of the subject.
        """
        if type(ps_dict) is not list:
            ps_dict = [ps_dict]

        for ps in ps_dict:
            self.ps.append(ps)

    def add_mri_info(self, mri_dict):
        """
        Adding MRI information to the subject object (multiple MRIs possible).

        Parameters
        ----------
        mri_dict : dict or list of dict
            Dictionary containing the MRI information of the subject.

        Notes
        -----
        **Adds Attributes**

        Subject.mri : list of dict
            Dictionary containing the MRI information of the subject.
        """
        if type(mri_dict) is not list:
            mri_dict = [mri_dict]

        for mri in mri_dict:
            self.mri.append(mri)

    @staticmethod
    def _prep_fn(val):
        if type(val) is dict:
            pass
        elif type(val) is not list:
            val = [val]
            for j in range(len(val)):
                _ = check_file_and_format(val[j])
        return val

    def add_experiment_info(self, exp_dict):
        """
        Adding information about a particular experiment.

        Parameters
        ----------
        exp_dict : dict of dict or list of dict
            Dictionary containing information about the experiment.

        Notes
        -----
        **Adds Attributes**

        exp : list of dict
            Dictionary containing information about the experiment.
        """
        fns = ['fn_data', 'fn_coil', 'fn_mri_nii']
        if exp_dict is None:
            return

        # if type(exp_dict) is not list:
        #     exp_dict = [exp_dict
        # check if files and folder exist and convert to list
        for i in exp_dict.keys():

            # # fn_tms_nav
            # if type(exp_dict[i]['fn_tms_nav']) is not list:
            #     exp_dict[i]['fn_tms_nav'] = [exp_dict[i]['fn_tms_nav']]
            # for j in range(len(exp_dict[i]['fn_tms_nav'])):
            #     _ = check_file_and_format(exp_dict[i]['fn_tms_nav'][j])

            # fn_data
            for fn in fns:
                try:
                    exp_dict[i][fn] = self._prep_fn(exp_dict[i][fn])
                except KeyError:
                    pass

            # new subject files have 'cond' as dictionary
            if 'cond' in exp_dict[i].keys():
                if isinstance(exp_dict[i]['cond'], dict):
                    for cond_name in exp_dict[i]['cond'].keys():
                        for fn in fns:
                            try:
                                exp_dict[i]['cond'][cond_name][fn] = self._prep_fn(exp_dict[i]['cond'][cond_name][fn])
                            except KeyError:
                                pass
            else:
                exp_dict[i]['cond'] = [[""]]

        for exp in exp_dict:
            self.exp[exp] = exp_dict[exp]


def fill_from_dict(obj, d):
    """
    Set all attributes from d in obj.

    Parameters
    ----------
    obj : pynibs.Mesh or pynibs.roi.ROI
        Object to fill with ``d``.
    d : dict
        Dictionary containing the attributes to set.

    Returns
    -------
    obj : pynibs.Mesh or pynibs.ROI
        Object with attributes set from ``d``.
    """
    for key, value in d.items():
        if not hasattr(obj, f"{key}"):
            warnings.warn(f"{key} not existing.")
        setattr(obj, key, value)

    return obj


def save_subject(subject_id, subject_folder, fname, mri_dict=None, mesh_dict=None, roi_dict=None,
                 exp_dict=None, ps_dict=None, **kwargs):
    """
    Saves subject information in .pkl or .hdf5 format (preferred)

    Parameters
    ----------
    subject_id : str
        ID of subject.
    subject_folder : str
        Subject folder
    fname : str
        Filename with .hdf5 or .pkl extension (incl. path).
    mri_dict : list of dict, optional
        MRI info.
    mesh_dict : list of dict, optional
        Mesh info.
    roi_dict : list of list of dict, optional
        Mesh info.
    exp_dict : list of dict, optional
        Experiment info.
    ps_dict : list of dict, optional
        Plot-settings info.
    kwargs : str or np.ndarray
        Additional information saved in the parent folder of the .hdf5 file.

    Returns
    -------
    <File> : .hdf5 file
        Subject information
    """
    filetype = os.path.splitext(fname)[1]

    if filetype == ".hdf5":
        if os.path.exists(fname):
            os.remove(fname)

        save_subject_hdf5(subject_id=subject_id,
                          subject_folder=subject_folder,
                          fname=fname,
                          mri_dict=mri_dict,
                          mesh_dict=mesh_dict,
                          roi_dict=roi_dict,
                          exp_dict=exp_dict,
                          ps_dict=ps_dict,
                          **kwargs)

    elif filetype == ".pkl":
        raise NotImplementedError
        #
        # # create and initialize subject
        # subject = Subject(subject_id=subject_id)
        #
        # # add mri information
        # subject.add_mri_info(mri_dict=mri_dict)
        #
        # # add mesh information
        # subject.add_mesh_info(mesh_dict=mesh_dict)
        #
        # # add roi info to mesh
        # subject.add_roi_info(roi_dict=roi_dict)
        #
        # # add experiment info
        # subject.add_experiment_info(exp_dict=exp_dict)
        #
        # # add plotsettings
        # subject.add_plotsettings(ps_dict=ps_dict)
        #
        # save_subject_pkl(sobj=subject, fname=fname)


def save_subject_pkl(sobj, fname):
    """
    Saving subject object as pickle file.

    Parameters
    ----------
    sobj: object
        Subject object to save
    fname: str
        Filename with .pkl extension

    Returns
    -------
    <File> : .pkl file
        Subject object instance
    """

    if type(fname) is list:
        fname = fname[0]

    with open(fname, 'wb') as output:
        pickle.dump(sobj, output, -1)


def save_subject_hdf5(subject_id, subject_folder, fname, mri_dict=None, mesh_dict=None, roi_dict=None,
                      exp_dict=None, ps_dict=None, overwrite=True, check_file_exist=False, verbose=False, **kwargs):
    """
    Saving subject information in hdf5 file.

    Parameters
    ----------
    subject_id : str
        ID of subject.
    subject_folder : str
        Subject folder.
    fname : str
        Filename with .hdf5 extension (incl. path).
    mri_dict : list of dict, optional
        MRI info.
    mesh_dict : list of dict, optional
        Mesh info.
    roi_dict : list of list of dict, optional
        Mesh info.
    exp_dict : list of dict or dict of dict, optional
        Experiment info.
    ps_dict : list of dict, optional, default:None
        Plot-settings info.
    overwrite : bool
        Overwrites existing .hdf5 file.
    check_file_exist : bool
        Hide warnings.
    verbose : bool
        Print information about meshes and ROIs.
    kwargs : str or np.ndarray
        Additional information saved in the parent folder of the .hdf5 file.

    Returns
    -------
    <File> : .hdf5 file
        Subject information.
    """
    assert fname.endswith('.hdf5')

    if overwrite and os.path.exists(fname):
        os.remove(fname)

    with h5py.File(fname, 'a') as f:
        f["subject_id"] = np.array(subject_id).astype("|S")
        f["subject_folder"] = np.array(subject_folder).astype("|S")

    if mri_dict is not None:
        if isinstance(mri_dict, list):
            mri_dict = {i: mri_dict[i] for i in range(len(mri_dict))}

        for i in mri_dict.keys():
            pynibs.write_dict_to_hdf5(fn_hdf5=fname, data=mri_dict[i], folder=f"mri/{i}", check_file_exist=True)

    if mesh_dict is not None:
        if isinstance(mesh_dict, list):
            mesh_dict = {i: mesh_dict[i] for i in range(len(mesh_dict))}

        for mesh_name, mesh_dict in mesh_dict.items():
            mesh = pynibs.Mesh(mesh_name=mesh_name, subject_id=subject_id, subject_folder=subject_folder)
            mesh.fill_defaults(mesh_dict['approach'])
            mesh = fill_from_dict(mesh, mesh_dict)
            mesh.write_to_hdf5(fn_hdf5=fname, check_file_exist=check_file_exist, verbose=verbose)

    if roi_dict is not None:
        for mesh_name in roi_dict.keys():
            for roi_name, roi_dict_i in roi_dict[mesh_name].items():
                roi = pynibs.ROI(subject_id=subject_id, roi_name=roi_name, mesh_name=mesh_name)
                roi = fill_from_dict(roi, roi_dict_i)
                roi.write_to_hdf5(fn_hdf5=fname, check_file_exist=check_file_exist, verbose=verbose)

    if exp_dict is not None:
        if isinstance(exp_dict, list):
            exp_dict = {i: exp_dict[i] for i in range(len(exp_dict))}
        for i in exp_dict.keys():
            pynibs.write_dict_to_hdf5(fn_hdf5=fname, data=exp_dict[i], folder=f"exp/{i}",
                                      check_file_exist=check_file_exist)

    if ps_dict is not None:
        for i in range(len(ps_dict)):
            pynibs.write_dict_to_hdf5(fn_hdf5=fname, data=ps_dict[i], folder=f"ps/{i}",
                                      check_file_exist=check_file_exist)

    with h5py.File(fname, 'a') as f:
        for key, value in kwargs.items():
            try:
                del f[key]
            except KeyError:
                pass
            f.create_dataset(name=key, data=value)


def load_subject_hdf5(fname):
    """
    Loading subject information from .hdf5 file and returning subject object.

    Parameters
    ----------
    fname : str
        Filename with .hdf5 extension (incl. path).

    Returns
    -------
    subject : pynibs.subject.Subject
        The Subject object.
    """
    with h5py.File(fname, 'r') as f:
        subject_id = str(f["subject_id"][()].astype(str))
        subject_folder = str(f["subject_folder"][()].astype(str))

        # create and initialize subject
        subject = Subject(subject_id=subject_id, subject_folder=subject_folder)

        # add mri information
        try:
            mri_keys = f["mri"].keys()
            mri = []

            for key in mri_keys:
                mri.append(pynibs.read_dict_from_hdf5(fn_hdf5=fname, folder=f"mri/{key}"))

        except KeyError:
            mri = None

        subject.add_mri_info(mri_dict=mri)

        # add mesh information
        try:
            mesh_keys = f["mesh"].keys()
            mesh = {}

            for key in mesh_keys:
                mesh[key] = pynibs.read_dict_from_hdf5(fn_hdf5=fname, folder=f"mesh/{key}")

        except KeyError:
            mesh = None

        subject.add_mesh_info(mesh_dict=mesh)

        # add roi info
        try:
            mesh_idx_keys = f["roi"].keys()
            roi = dict()

            for mesh_idx in mesh_idx_keys:
                try:
                    roi_idx_keys = f[f"roi/{mesh_idx}"].keys()
                    roi[mesh_idx] = dict()

                    for roi_idx in roi_idx_keys:
                        roi[mesh_idx][roi_idx] = pynibs.read_dict_from_hdf5(fn_hdf5=fname,
                                                                            folder=f"roi/{mesh_idx}/{roi_idx}")

                except KeyError:
                    roi[mesh_idx] = None

            subject.add_roi_info(roi_dict=roi)

        except KeyError:
            pass

        # add experiment information
        try:
            exp_keys = f["exp"].keys()
            exp = {}

            for key in exp_keys:
                exp[key] = pynibs.read_dict_from_hdf5(fn_hdf5=fname, folder=f"exp/{key}")

        except KeyError:
            exp = None

        subject.add_experiment_info(exp_dict=exp)

        # add plotsettings information
        try:
            ps_keys = f["ps"].keys()
            ps = []

            for key in ps_keys:
                ps.append(pynibs.read_dict_from_hdf5(fn_hdf5=fname, folder=f"ps/{key}"))

        except KeyError:
            ps = None

        subject.add_plotsettings(ps_dict=ps)

    return subject


def load_subject(fname, filetype=None):
    """
    Wrapper for pkl and hdf5 subject loader

    Parameters
    ----------
    fname: str
        endswith('.pkl') | endswith('.hdf5').
    filetype: str
        Explicitly set file version.

    Returns
    -------
    subject : pynibs.subject.Subject
        Loaded Subject object.
    """
    # explicit set fname type
    if filetype:
        if filetype.lower().endswith("hdf5"):
            filetype = "hdf5"
        elif filetype.lower().endswith("pkl"):
            filetype = "pkl"
        else:
            raise NotImplementedError(f"{filetype} unknown.")

    # determine fname type from file-ending
    else:
        if fname.lower().endswith("hdf5"):
            filetype = "hdf5"
        elif fname.lower().endswith("pkl"):
            filetype = "pkl"
        else:
            raise NotImplementedError(f"{fname} type unknown.")

    # load file using correct load function
    if filetype == "hdf5":
        return load_subject_hdf5(fname)
    elif filetype == "pkl":
        return load_subject_pkl(fname)


def load_subject_pkl(fname):
    """
    Loading subject object from .pkl file.

    Parameters
    ----------
    fname : str
        Filename with .pkl extension.

    Returns
    -------
    subject : pynibs.subject.Subject
        Loaded Subject object.
    """
    try:
        with open(fname, 'rb') as f:
            return pickle.load(f)

    except UnicodeDecodeError:
        print(".pkl file version does not match python version... recreating subject object")
        fn_scipt = os.path.join(os.path.split(fname)[0],
                                "create_subject_" +
                                os.path.splitext(os.path.split(fname)[1])[0] + ".py")
        pynibs.bash_call("python {}".format(fn_scipt))

        with open(fname, 'rb') as f:
            return pickle.load(f)


def check_file_and_format(fname):
    """
    Checking existence of file and transforming to list if necessary.

    Parameters
    ----------
    fname: str or list of str
        Filename(s) to check.

    Returns
    -------
    fname: list of str
        Checked filename(s) as list.
    """
    if type(fname) is not list:
        fname = [fname]

    for fn in fname:
        if not (os.path.exists(fn)):
            Exception('File/Folder {} does not exist!'.format(fn))

    return fname


def create_plot_settings_dict(plotfunction_type):
    """
    Creates a dictionary with default plotsettings.

    Parameters
    ----------
    plotfunction_type : str
        Plot function the dictionary is generated for:

        * 'surface_vector_plot'
        * 'surface_vector_plot_vtu'
        * 'volume_plot'
        * 'volume_plot_vtu'

    Returns
    -------
    ps : dict
        Dictionary containing the plotsettings.
    axes : bool
        Show orientation axes.
    background_color : nparray
         (1m 3) Set background color of exported image RGB (0...1).
    calculator : str
        Format string with placeholder of the calculator expression the quantity to plot is modified with,
        e.g.: "{}^5".
    clip_coords : nparray of float
        (N_clips, 3) Coordinates of clip surface origins (x,y,z).
    clip_normals : nparray of float
        (N_clips, 3) Surface normals of clip surfaces pointing in the direction where the volume is kept for
        clip_type = ['clip' ...] (x,y,z).
    clip_type : list of str
        Type of clipping:

        * 'clip':  cut geometry but keep volume behind
        * 'slice': cut geometry and keep only the slice
    coil_dipole_scaling : list [1 x 2]
        Specify the scaling type of the dipoles (2 entries):
        ``coil_dipole_scaling[0]``:

        * 'uniform': uniform scaling, i.e. all dipoles have the same size
        + 'scaled': size scaled according to dipole magnitude

        ``coil_dipole_scaling[1]``:

        * scalar scale parameter of dipole size
    coil_dipole_color : str or list
        Color of the dipoles; either str to specify colormap (e.g. 'jet') or list of RGB values [1 x 3] (0...1).
    coil_axes : bool, default: True
        Plot coil axes visualizing the principle direction and orientation of the coil.
    colorbar_label : str
        Label of plotted data close to colorbar.
    colorbar_position : list of float
        (1, 2) Position of colorbar (lower left corner) 0...1 [x_pos, y_pos].
    colorbar_orientation : str
        Orientation of colorbar (``'Vertical'``, ``'Horizontal'``).
    colorbar_aspectratio : int
        Aspectratio of colorbar (higher values make it thicker).
    colorbar_titlefontsize : float
        Fontsize of colorbar title.
    colorbar_labelfontsize : float
        Fontsize of colorbar labels (numbers).
    colorbar_labelformat : str
        Format of colorbar labels (e.g.: '%-#6.3g').
    colorbar_numberoflabels : int
        maximum number of colorbar labels.
    colorbar_labelcolor : list of float
        (1, 3) Color of colorbar labels in RGB (0...1).
    colormap : str or nparray
        If nparray [1 x 4*N]: custom colormap providing data and corresponding RGB values

        .. math::
            \\begin{bmatrix}
              data_{1} & R_1 & G_1 & B_1  \\\\
              data_{2} & R_2 & G_2 & B_2  \\\\
              ...      & ... & ... & ...  \\\\
              data_{N} & R_N & G_N & B_N  \\\\
             \\end{bmatrix}

        if str: colormap of plotted data chosen from included presets:

        * 'Cool to Warm',
        * 'Cool to Warm (Extended)',
        * 'Blue to Red Rainbow',
        * 'X Ray',
        * 'Grayscale',
        * 'jet',
        * 'hsv',
        * 'erdc_iceFire_L',
        * 'Plasma (matplotlib)',
        * 'Viridis (matplotlib)',
        * 'gray_Matlab',
        * 'Spectral_lowBlue',
        + 'BuRd'
        * 'Rainbow Blended White'
        * 'b2rcw'
    colormap_categories : bool
        Use categorized (discrete) colormap.
    datarange : list
        (1, 2) Minimum and Maximum of plotted datarange [MIN, MAX] (default: automatic).
    domain_IDs : int or list of int
        Domain IDs
        surface plot: Index of surface where the data is plotted on (Default: 0)
        volume plot: Specify the domains IDs to show in plot (default: all) Attention! Has to be included in the
        dataset under the name 'tissue'! e.g. for SimNIBS:

        * 1 -> white matter (WM)
        * 2 -> grey matter (GM)
        * 3 -> cerebrospinal fluid (CSF)
        * 4 -> skull
        * 5 -> skin
    domain_label : str
        Label of the dataset which contains the domain IDs (default: 'tissue_type').
    edges : BOOL
        Show edges of mesh.
    fname_in : str or list of str
        Filenames of input files, 2 possibilities:

        * .xdmf-file: filename of .xmdf (needs the corresponding .hdf5 file(s) in the same folder)
        * .hdf5-file(s): filename(s) of .hdf5 file(s) containing the data and the geometry. The data can be provided
          in the first hdf5 file and the geometry can be provided in the second file. However, both can be also
          provided in a single hdf5 file.
    fname_png : str
        Name of output .png file (incl. path).
    fname_vtu_volume : str
        Name of .vtu volume file containing volume data (incl. path).
    fname_vtu_surface : str
        Name of .vtu surface file containing surface data (incl. path) (to distinguish tissues).
    fname_vtu_coil : str, optional
        Name of coil .vtu file (incl. path).
    info : str
        Information about the plot the settings belong to.
    interpolate : bool
        Interpolate data for visual smoothness.
    NanColor : list of float
        (3) RGB color values for "Not a Number" values (range 0 ... 1).
    opacitymap : np.ndarray
        Points defining the piecewise linear opacity transfer function (transparency) (default: no transparency)
        connecting data values with opacity (alpha) values ranging from 0 (max. transparency) to 1 (no transparency).

        .. math::
           \\begin{bmatrix}
            data_{1} & opac_1 & 0.5 & 0  \\\\
            data_{2} & opac_2 & 0.5 & 0  \\\\
            ...      & ...   & ... & ...\\\\
            data_{N} & opac_N & 0.5 & 0  \\\\
           \\end{bmatrix}
    plot_function : str
        Function the plot is generated with:

        - 'surface_vector_plot'
        - 'surface_vector_plot_vtu'
        - 'volume_plot'
        - 'volume_plot_vtu'
    png_resolution : float
        Resolution parameter of output image (1...5).
    quantity : str
        Label of magnitude dataset to plot.
    surface_color : np.ndarray
        (1, 3) Color of brain surface in RGB (0...1) for better visibility of tissue borders.
    surface_smoothing : bool
        Smooth the plotted surface (True/False).
    show_coil : bool, default: True
        show coil if present in dataset as block termed 'coil'.
    vcolor : np.ndarray of float
        (N_vecs, 3) Array containing the RGB values between 0...1 of the vector groups in dataset to plot.
    vector_mode : dict
        Key determines the type how many vectors are shown:

        - 'All Points'
        - 'Every Nth Point'
        - 'Uniform Spatial Distribution'

        Value (int) is the corresponding number of vectors:

        - 'All Points' (not set)
        - 'Every Nth Point' (every Nth vector is shown in the grid)
        - 'Uniform Spatial Distribution' (not set)
    view : list
        Camera position and angle in 3D space:
        ``[[3 x CameraPosition], [3 x CameraFocalPoint], [3 x CameraViewUp], 1 x CameraParallelScale]``.
    viewsize : nparray [1 x 2]
        Set size of exported image in pixel [width x height] will be extra scaled by parameter png_resolution.
    vlabels : list of str
        Labels of vector datasets to plot (other present datasets are ignored).
    vscales : list of float
        Scale parameters of vector groups to plot.
    vscale_mode : list of str [N_vecs x 1]
        List containing the type of vector scaling:

        - 'off': all vectors are normalized
        - 'vector': vectors are scaled according to their magnitudeeee
    """
    if plotfunction_type not in ['surface_vector_plot', 'surface_vector_plot_vtu', 'volume_plot', 'volume_plot_vtu']:
        raise Exception('plotfunction_type not set correctly, specify either [\'surface_vector_plot\', \
                            \'surface_vector_plot_vtu\', \'volume_plot\', \'volume_plot_vtu\']')

    ps = dict()
    ps['info'] = []
    ps['plot_function'] = plotfunction_type
    ps['fname_png'] = []
    ps['png_resolution'] = 1
    ps['quantity'] = None
    ps['datarange'] = None
    ps['colorbar_label'] = ''
    ps['colorbar_position'] = []
    ps['colorbar_orientation'] = 'Vertical'
    ps['colorbar_aspectratio'] = 20
    ps['colorbar_titlefontsize'] = 7
    ps['colorbar_labelfontsize'] = 7
    ps['colorbar_labelformat'] = '%-#6.3g'
    ps['colorbar_numberoflabels'] = 5
    ps['colorbar_labelcolor'] = [0, 0, 0]
    ps['colorbar_font'] = 'Arial'
    ps['view'] = 0
    ps['interpolate'] = False
    ps['edges'] = False
    ps['axes'] = True
    ps['colormap'] = "jet"
    ps['colormap_categories'] = False
    ps['opacitymap'] = None
    ps['background_color'] = [1, 1, 1]
    ps['viewsize'] = [800, 800]
    ps['NanColor'] = [1, 1, 1]
    ps["surface_smoothing"] = True
    ps["calculator"] = None

    # filename of input files
    if plotfunction_type in ['surface_vector_plot', 'surface_vector_plot_vtu', 'volume_plot']:
        ps['fname_in'] = []

    elif plotfunction_type in ['volume_plot_vtu']:
        ps['fname_vtu_volume'] = []
        ps['fname_vtu_surface'] = []
        ps['fname_vtu_coil'] = []

    # new features in .xdmf plot functions
    if plotfunction_type in ['surface_vector_plot', 'volume_plot', 'volume_plot_vtu']:
        ps['domain_label'] = 'tissue_type'
        ps['domain_IDs'] = 0
        ps['show_coil'] = True
        ps['coil_dipole_scaling'] = ['uniform', 1]
        ps['coil_dipole_color'] = 'jet'
        ps['coil_axes'] = True

    # surface specific properties
    if plotfunction_type in ['surface_vector_plot', 'surface_vector_plot_vtu']:
        ps['vlabels'] = ''
        ps['vscales'] = np.array([1])
        ps['vscale_mode'] = 'vector'
        ps['vector_mode'] = {'All Points': 0}
        ps['vcolor'] = np.array([1.0, 0.0, 0.0])

    # volume specific properties
    if plotfunction_type in ['volume_plot', 'volume_plot_vtu']:
        ps['clip_coords'] = np.array([])
        ps['clip_normals'] = np.array([])
        ps['clip_type'] = []
        ps['surface_color'] = [1, 1, 1]

    return ps



""" """
import os
import h5py
import numpy as np
import pynibs


def cfs2hdf5(fn_cfs, fn_hdf5=None):
    """
    Converts EMG data included in .cfs file to .hdf5 format.

    Parameters
    ----------
    fn_cfs : str
        Filename of .cfs file.
    fn_hdf5 : str, optional
        Filename of .hdf5 file (if not provided, a file with same name as fn_cfs will be created with .hdf5 extension).

    Returns
    -------
    <file> : .hdf5 File
        File containing:

        * EMG data in f["emg"][:]
        * Time axis in f["time"][:]
    """
    try:
        import biosig
    except ImportError:
        ImportError("Please install biosig from pynibs/pkg/biosig folder!")

    if fn_hdf5 is None:
        fn_hdf5 = os.path.splitext(fn_cfs)[0] + ".hdf5"

    # load header and data
    cfs_header = biosig.header(fn_cfs)
    emg = biosig.data(fn_cfs)[:, 0]

    sweep_index = cfs_header.find('NumberOfSweeps')
    comma_index = cfs_header.find(',', sweep_index)
    sweeps = int(cfs_header[sweep_index + 18:comma_index])
    records = emg.shape[0]
    samples = int(records / sweeps)
    sampling_rate = pynibs.get_mep_sampling_rate(fn_cfs)
    emg = np.reshape(emg, (sweeps, samples))
    time = np.linspace(0, samples, samples) / sampling_rate

    with h5py.File(fn_hdf5, "w") as f:
        f.create_dataset("emg", data=emg)
        f.create_dataset("time", data=time)
        f.create_dataset("sampling_rate", data=np.array([sampling_rate]))



"""
Functions to optimize single coil currents for multichannel TMS arrays.
"""
import numpy as np
from scipy.optimize import minimize


def get_score_raw(x, e, n_stim, n_ele, n_channel, x_opt=None, opt_target='elms'):
    """
    Compute score for e-efield cross correlations.
    Non-normalized score is returned, so you need to do sth like

    score = 1 / ((n_ele ** 2 - n_ele) / 2) * get_score()

    Parameters
    ----------
    x : np.ndarray of float
        (n_channel * n_stim_opt, ) Vector to scale channels for each.
    e : np.ndarray of float
        (n_ele*3, n_channels) E-field.
    n_stim : int
        Number of stimulations to compute score for.
    n_ele : int
        Number of elements.
    n_channel : int
        Number of channels.
    x_opt :  np.ndarray of float, optinonal
        (n_pre_opt,) Previously optimized channel currents.
    opt_target : str, default: 'elms'
        Optimization target. 'elms' for optimizing decorrelations of elements, 'stims' for stimulations.

    Returns
    -------
    score : float
        non-normalized score np.nansum(np.abs(np.triu(np.corrcoef(e_mag), k=1)))
    """
    if x_opt is None:
        currents_all_zaps_channels = np.reshape(x, (n_channel, n_stim))
    else:
        currents_all_zaps_channels = np.hstack((x_opt, np.reshape(x, (n_channel, n_stim - x_opt.shape[1]))))
    # currents_all_zaps_channels.shape = (n_chans, n_zaps)

    # determine total electric field (vector form)
    e_vec = np.matmul(e, currents_all_zaps_channels)  # e_vec.shape = (n_elems*3, n_zaps)
    # determine magnitude
    # for e_ in e_vec.T:
    #     a = np.linalg.norm(np.reshape(e_, (n_ele, 3)), axis=1)
    #     # e_.shape = (n_ele=3,)
    #     # np.reshape(e_, (n_ele, 3)).shape = (n_ele,3)
    #     # a.shape = (n_ele,)

    if e.ndim == 2:
        e_mag = np.vstack([np.linalg.norm(np.reshape(e_, (n_ele, 3)), axis=1) for e_ in e_vec.T]).T
    elif e.ndim == 3:
        e_mag = np.linalg.norm(e_vec, axis=1)
    else:
        raise ValueError
    # e_mag.shape = (n_ele, n_zaps)

    # determine average correlation coefficient
    # r_avg = 1/((n_ele**2-n_ele)/2) * np.sum(np.abs(np.triu(np.corrcoef(e_mag), k=1)))
    if opt_target == 'elms':
        r_sum = np.nansum(np.abs(np.triu(np.corrcoef(e_mag), k=1)))
    elif opt_target == 'stims':
        r_sum = np.nansum(np.abs(np.triu(np.corrcoef(e_mag.T), k=1)))
    else:
        raise ValueError("opt_target has to be 'elms' or 'stims'")

    return r_sum


def get_score_raw_single_channel(x, e, x_opt=None):
    """
    Compute score for e-efield cross correlations.
    Non-normalized score is returned, so you need to do sth like

    score = 1 / ((n_ele ** 2 - n_ele) / 2) * get_score()

    Parameters
    ----------
    x : np.ndarray of float
        (n_placements, ) selection of coil placements.
    e : np.ndarray of float
        (n_ele*3, n_channels) E-field.
    x_opt :  np.ndarray of float, optinonal
        (n_pre_opt,) Previously optimized channel currents.

    Returns
    -------
    score : float
        non-normalized score np.nansum(np.abs(np.triu(np.corrcoef(e_mag), k=1)))
    """
    # x = ((x / x.max()) * e.shape[0]-1).astype(int)
    if x_opt is not None:
        x = np.hstack((x_opt, x))
    x = x.astype(int)
    e_mag = e[x, :]

    # determine average correlation coefficient
    # r_avg = 1/((n_ele**2-n_ele)/2) * np.sum(np.abs(np.triu(np.corrcoef(e_mag), k=1)))
    # a = np.abs(np.triu(np.corrcoef(e_mag.T), k=1))

    r_avg = np.nansum(np.abs(np.triu(np.corrcoef(e_mag.T), k=1)))
    return r_avg


def optimize_currents(e, n_stim, currents_prev=None, seed=None,
                      maxiter=200, method='SLSQP', opt_target='elms', verbose=False):
    """
    Optimize the currents for a multichannel TMS array by minimizing e-fields cross-correlation.

    Parameters
    ----------
    e : np.ndarray of float
        (n_elms * 3, n_channel) or (n_elms, 3, n_channel). E in ROI for currents = 1.
    n_stim : int
        Number of stimulations.
    currents_prev : np.ndarray of float, optional
        (n_channels, n_stims_prev) Previous currents to append to.
    seed : int, optional
        Seed for random number generator.
    maxiter : int, default=200
        Max iterations of the optimization.
    method : str, default: 'SLSQP'
        Optimization method.
    verbose : bool, default: False
        Print additional information.
    opt_target : str, default: 'elms'
        Optimization target. 'elms' for optimizing decorrelations of elements, 'stims' for stimulations.

    Returns
    -------
    currents : np.ndarray
        (n_channels, n_stims) The optimized currents to drive the multichannel array.
    score : float
        Final score of the solution.
    """
    if e.ndim == 2:
        n_channel = e.shape[1]
        n_ele = int(e.shape[0] / 3)
    elif e.ndim == 3:
        n_channel = e.shape[2]
        n_ele = int(e.shape[0])
    else:
        raise ValueError

    if currents_prev is None:
        n_stim_opt = n_stim
    else:
        n_stim_opt = n_stim - currents_prev.shape[1]

        if n_stim_opt <= 0:
            raise ValueError("N_stim has to be larger than already optimized optimal values!")

    # initial guess for currents for all channels * stimulations
    if seed is not None:
        np.random.seed(seed)
    x0 = (np.random.rand(n_channel * n_stim_opt) * 2) -1
    # print(x0[:5])
    if verbose:
        print(f"n_ele: {n_ele}, n_channels: {n_channel}, n_stims: {n_stim}")

    # optimization algorithm
    res = minimize(get_score_raw,
                   args=(e, n_stim, n_ele, n_channel, currents_prev, opt_target),
                   x0=x0,
                   method=method,
                   options={'disp': False, 'maxiter': maxiter},
                   bounds=[(-1, 1) for _ in range(len(x0))],
                   tol=1e-6)
    # print(res.fun, res.success, res.message)

    if currents_prev is None:
        currents = np.reshape(res.x, (n_channel, n_stim))
    else:
        currents = np.hstack((currents_prev, np.reshape(res.x, (n_channel, n_stim - currents_prev.shape[1]))))
    if opt_target == 'elms':
        score = 1 / ((n_ele ** 2 - n_ele) / 2) * res.fun
    elif opt_target == 'stims':
        score = 1 / ((n_stim ** 2 - n_stim) / 2) * res.fun
    else:
        raise ValueError("opt_target has to be 'elms' or 'stims'")

    return currents, score, res


def optimize_currents_single_channel(e, n_stim, currents_prev=None, seed=None,
                                     maxiter=200, method='SLSQP', verbose=False):
    """
    Optimize the coil placement selection for a single channel e-field set minimizing e-fields cross-correlation.

    Parameters
    ----------
    e : np.ndarray of float
        (n_elms3, n_placements).
    n_stim : int
        Number of stimulations.
    currents_prev : np.ndarray of float, optional
        (n_channels, n_stims_prev) Previous currents to append to.
    seed : int, optional
        Seed for random number generator.
    maxiter : int, default=200
        Max iterations of the optimization.
    method : str, default: 'SLSQP'
        Optimization method.
    verbose : bool, default: False
        Print additional information.

    Returns
    -------
    currents :  np.ndarray
        (n_channels, n_stims) The optimized currents to drive the multichannel array.
    score : float
        Final score of the solution.
    """
    n_placements = e.shape[0]
    n_ele = e.shape[1]

    if currents_prev is None:
        n_stim_opt = n_stim
    else:
        n_stim_opt = n_stim - currents_prev.shape[0]

        if n_stim_opt <= 0:
            raise ValueError("N_stim has to be larger than already optimized optimal values!")

    # initial guess for currents for all channels * stimulations
    if seed is not None:
        np.random.seed(seed)
    x0 = np.random.randint(0, n_placements, n_stim_opt)
    if verbose:
        print(f"n_ele: {n_ele}, n_placements: {n_placements}, n_stims: {n_stim}")

    # optimization algorithm
    res = minimize(get_score_raw_single_channel,
                   args=(e, currents_prev),
                   x0=x0,
                   method=method,
                   options={'disp': False, 'maxiter': maxiter},
                   bounds=[(0, n_placements) for _ in range(len(x0))])

    # if currents_prev is None:
    #     currents = np.reshape(res.x, (n_placements, n_stim))
    # else:
    #     currents = np.hstack((currents_prev, np.reshape(res.x, (n_placements, n_stim - currents_prev.shape[1]))))

    score = 1 / ((n_ele ** 2 - n_ele) / 2) * res.fun
    # score = res.fun

    return res.x, score


def get_score(x, e, n_stim, n_ele, n_channel, x_opt=None):
    """
    Normalize the score by the number of elements.

    Parameters
    ----------
    x : np.ndarray of float
        (n_channel * n_stim_opt, ) Vector to scale channels for each.
    e : np.ndarray of float
        (n_ele*3, n_channels) E-field.
    n_stim : int
        Number of stimulations to compute score for.
    n_ele : int
        Number of elements.
    n_channel : int
        Number of channels.
    x_opt :  np.ndarray of float, optinonal
        (n_pre_opt,) Previously optimized channel currents.

    Returns
    -------
    score : float
        The normalized score.
    """
    score_raw = get_score_raw(x, e, n_stim, n_ele, n_channel, x_opt)
    return 1 / ((n_ele ** 2 - n_ele) / 2) * score_raw



import unittest
import numpy as np
import os
import meshio
import tempfile
from pynibs import (point_data_to_cell_data_vtk,
                    cell_data_to_point_data_vtk,
                    cell_data_to_point_data,
                    data_nodes2elements)


class TestMeshFunctions(unittest.TestCase):
    def setUp(self):
        # Create a simple mesh for testing
        self.points = np.array([[0, 0, 0], [1, 0, 0], [0, 1, 0]])
        self.con = np.array([[0, 1, 2]])
        cells = [("triangle", self.con)]

        self.cell_data = {"test": [np.array([[1]])]}
        self.mesh_cell_data_1d = meshio.Mesh(self.points, cells, cell_data=self.cell_data)

        self.cell_data = {"test": [np.array([[1, 2, 3]])]}
        self.mesh_cell_data_3d = meshio.Mesh(self.points, cells, cell_data=self.cell_data)

        cell_data = {"test": [np.array([[1, 2, 3]])],
                     "test2": [np.array([[100, 200, 300]])]}
        self.mesh_cell_datas_3d = meshio.Mesh(self.points, cells, cell_data=cell_data)

        # Create another mesh for testing with point data
        self.point_data = {"test": np.array([[1, 2, 3],
                                             [1, 2, 3],
                                             [1, 2, 3]])}
        self.mesh_point_data_3d = meshio.Mesh(self.points, cells, point_data=self.point_data)

        point_data = {"test": np.array([1, 2, 3])}
        self.mesh_point_data_1d = meshio.Mesh(self.points, cells, point_data=point_data)

        point_data = {"test": np.array([1, 2, 3]),
                      "test2": np.array([100, 200, 300])}
        self.mesh_point_datas_1d = meshio.Mesh(self.points, cells, point_data=point_data)

    def test_point_to_cell_data(self):
        # Test conversion from point data to cell data for both meshes
        for mesh in [self.mesh_point_data_1d, self.mesh_point_data_3d]:
            with tempfile.NamedTemporaryFile(suffix=".vtu") as temp_file:
                cell_data = point_data_to_cell_data_vtk(mesh, fn=temp_file.name)
                self.assertTrue("test" in cell_data)
                self.assertEqual(cell_data["test"].shape[0], mesh.cells[0].data.shape[0])
                self.assertTrue(os.path.exists(temp_file.name))

    def test_cell_to_point_data(self):
        # Test conversion from cell data to point data for both meshes
        for mesh in [self.mesh_cell_data_1d, self.mesh_cell_data_3d]:
            with tempfile.NamedTemporaryFile(suffix=".vtu") as temp_file:
                point_data = cell_data_to_point_data_vtk(mesh, fn=temp_file.name)
                self.assertTrue("test" in point_data)
                self.assertEqual(point_data["test"].shape[0], mesh.points.shape[0])
                self.assertTrue(os.path.exists(temp_file.name))

    def test_point_to_cell_data_from_con(self):
        # Test conversion from point data to cell data for both meshes
        for mesh in [self.mesh_point_data_1d, self.mesh_point_data_3d]:
            with tempfile.NamedTemporaryFile(suffix=".vtu") as temp_file:
                cell_data = point_data_to_cell_data_vtk(nodes=self.points,
                                                        con=self.con,
                                                        point_data=self.point_data, fn=temp_file.name)
                self.assertTrue("test" in cell_data)
                self.assertEqual(cell_data["test"].shape[0], mesh.cells[0].data.shape[0])
                self.assertTrue(os.path.exists(temp_file.name))

    def test_cell_to_point_data_from_con(self):
        # Test conversion from cell data to point data for both meshes
        for mesh in [self.mesh_cell_data_1d, self.mesh_cell_data_3d]:
            with tempfile.NamedTemporaryFile(suffix=".vtu") as temp_file:
                point_data = cell_data_to_point_data_vtk(nodes=self.points,
                                                         con=self.con,
                                                         cell_data=self.cell_data,
                                                         fn=temp_file.name)
                self.assertTrue("test" in point_data)
                self.assertEqual(point_data["test"].shape[0], mesh.points.shape[0])
                self.assertTrue(os.path.exists(temp_file.name))

    def test_point_to_cell_multiple_data(self):
        # Test conversion from point data to cell data for both meshes
        mesh = self.mesh_point_datas_1d
        with tempfile.NamedTemporaryFile(suffix=".vtu") as temp_file:
            cell_data = point_data_to_cell_data_vtk(mesh, fn=temp_file.name)
            self.assertTrue("test" in cell_data)
            self.assertEqual(cell_data["test"].shape[0], mesh.cells[0].data.shape[0])
            self.assertTrue("test2" in cell_data)
            self.assertEqual(cell_data["test2"].shape[0], mesh.cells[0].data.shape[0])
            self.assertTrue(os.path.exists(temp_file.name))

    def test_cell_to_point_multiple_data(self):
        # Test conversion from cell data to point data for both meshes
        mesh = self.mesh_cell_datas_3d
        with tempfile.NamedTemporaryFile(suffix=".vtu") as temp_file:
            point_data = cell_data_to_point_data_vtk(mesh, fn=temp_file.name)
            self.assertTrue("test" in point_data)
            self.assertEqual(point_data["test"].shape[0], mesh.points.shape[0])
            self.assertTrue("test2" in point_data)
            self.assertEqual(point_data["test2"].shape[0], mesh.points.shape[0])
            self.assertTrue(os.path.exists(temp_file.name))

    def test_cell_data_to_point_data(self):
        # Convert the cell data to point data
        data_nodes = cell_data_to_point_data(self.con, self.cell_data['test'][0], self.points)

        # Check the result
        self.assertEqual(data_nodes.shape, self.points.shape)
        self.assertTrue(np.all(data_nodes >= 1) and np.all(data_nodes <= 4))

    def test_data_nodes2elements(self):
        # Convert the node data to element data
        data_tris = data_nodes2elements(self.point_data['test'], self.con)

        # Check the result
        self.assertEqual(data_tris.shape, self.con.shape)
        self.assertTrue(np.all(data_tris >= 1) and np.all(data_tris <= 4))


if __name__ == "__main__":
    unittest.main()



import os
import shutil
import tempfile

import pynibs
import unittest
from xml.sax import make_parser
from xml.sax.handler import ContentHandler


def check_xml(xml_fn):
    """Check well-formedness of xml"""
    parser = make_parser()
    parser.setContentHandler(ContentHandler())
    parser.parse(xml_fn)


class TestWriteXdmf(unittest.TestCase):
    """This tests pynibs.write_xdmf()"""

    data_fn = os.path.join(pynibs.__testdatadir__, "data.hdf5")
    geo_fn = os.path.join(pynibs.__testdatadir__, "geo.hdf5")
    xml_scheme_fn = os.path.join(pynibs.__testdatadir__, "vtk-unstructured.xsd")
    xdmf_geo_fn = os.path.join(pynibs.__testdatadir__, "geo.xdmf")
    xdmf_data_fn = os.path.join(pynibs.__testdatadir__, "data.xdmf")

    for fn in [xdmf_geo_fn, xdmf_data_fn]:
        try:
            os.remove(fn)
        except OSError:
            pass

    def test_1_write_geo_only(self):
        with tempfile.TemporaryDirectory() as tempdir:
            geo_fn = shutil.copy(self.geo_fn, tempdir)
            fn_xml = pynibs.write_xdmf(geo_fn)
            check_xml(fn_xml)

    def test_2_write_xdmf_exists(self):
        with tempfile.TemporaryDirectory() as tempdir:
            geo_fn = shutil.copy(self.geo_fn, tempdir)
            pynibs.write_xdmf(geo_fn)
            self.assertRaises(FileExistsError, pynibs.write_xdmf, geo_fn)

    def test_3_write_xdmf_overwrite(self):
        with tempfile.TemporaryDirectory() as tempdir:
            geo_fn = shutil.copy(self.geo_fn, tempdir)
            pynibs.write_xdmf(geo_fn)
            pynibs.write_xdmf(geo_fn, overwrite_xdmf=True)

    def test_4_write_datageo(self):
        with tempfile.TemporaryDirectory() as tempdir:
            geo_fn = shutil.copy(self.geo_fn, tempdir)
            data_fn = shutil.copy(self.data_fn, tempdir)
            pynibs.write_xdmf(geo_fn)
            xdmf_data_fn = pynibs.write_xdmf(data_fn, geo_fn, overwrite_xdmf=True)
            check_xml(xdmf_data_fn)


if __name__ == '__main__':
    unittest.main()



"""Some helper functions to take care of geometric rotations"""
import math
import numpy as np
from scipy.spatial.transform import Rotation


def normalize_rot(rot):
    """
    Normalize rotation matrix.

    Parameters
    ----------
    rot : np.ndarray of float
        (3, 3) Rotation matrix.

    Returns
    -------
    rot_norm : np.ndarray of float
        (3, 3) Normalized rotation matrix.
    """
    q = rot_to_quat(rot)
    q /= np.sqrt(np.sum(q ** 2))
    return quat_to_rot(q)


def quat_rotation_angle(q):
    """
    Computes the rotation angle from the quaternion in rad.

    Parameters
    ----------
    q : np.ndarray of float
        Quaternion, either only the imaginary part (length=3) [qx, qy, qz]
        or the full quaternion (length=4) [qw, qx, qy, qz].

    Returns
    -------
    alpha : float
        Rotation angle of quaternion in rad.
    """

    if len(q) == 3:
        return 2 * np.arcsin(np.linalg.norm(q))
    elif len(q) == 4:
        return q[0]
    else:
        raise ValueError('Please check size of quaternion')


def quat_to_rot(q):
    """
    Computes the rotation matrix from quaternions.

    Parameters
    ----------
    q : np.ndarray of float
        Quaternion, either only the imaginary part (length=3) or the full quaternion (length=4).

    Returns
    -------
    rot : np.ndarray of float
        (3, 3) Rotation matrix, containing the x, y, z axis in the columns.
    """
    if q.size == 3:
        q = np.hstack([np.sqrt(1 - np.sum(q ** 2)), q])
    rot = np.array([[q[0] ** 2 + q[1] ** 2 - q[2] ** 2 - q[3] ** 2, 2 * (q[1] * q[2] - q[0] * q[3]),
                     2 * (q[1] * q[3] + q[0] * q[2])],
                    [2 * (q[2] * q[1] + q[0] * q[3]), q[0] ** 2 - q[1] ** 2 + q[2] ** 2 - q[3] ** 2,
                     2 * (q[2] * q[3] - q[0] * q[1])],
                    [2 * (q[3] * q[1] - q[0] * q[2]), 2 * (q[3] * q[2] + q[0] * q[1]),
                     q[0] ** 2 - q[1] ** 2 - q[2] ** 2 + q[3] ** 2]])
    return rot


def rot_to_quat(rot):
    """
    Computes the quaternions from rotation matrix
    (see e.g. https://www.euclideanspace.com/maths/geometry/rotations/conversions/matrixToQuaternion/).

    Parameters
    ----------
    rot : np.ndarray of float
        (3, 3) Rotation matrix, containing the x, y, z axis in the columns.

    Returns
    -------
    q : np.ndarray of float
        Quaternion, full (length=4).
    """

    rot = rot.flatten()
    t = 1. + rot[0] + rot[4] + rot[8]
    if t > np.finfo(rot.dtype).eps:
        s = np.sqrt(t) * 2.
        qx = (rot[7] - rot[5]) / s
        qy = (rot[2] - rot[6]) / s
        qz = (rot[3] - rot[1]) / s
        qw = 0.25 * s
    elif rot[0] > rot[4] and rot[0] > rot[8]:
        s = np.sqrt(1. + rot[0] - rot[4] - rot[8]) * 2.
        qx = 0.25 * s
        qy = (rot[1] + rot[3]) / s
        qz = (rot[2] + rot[6]) / s
        qw = (rot[7] - rot[5]) / s
    elif rot[4] > rot[8]:
        s = np.sqrt(1. - rot[0] + rot[4] - rot[8]) * 2
        qx = (rot[1] + rot[3]) / s
        qy = 0.25 * s
        qz = (rot[5] + rot[7]) / s
        qw = (rot[2] - rot[6]) / s
    else:
        s = np.sqrt(1. - rot[0] - rot[4] + rot[8]) * 2.
        qx = (rot[2] + rot[6]) / s
        qy = (rot[5] + rot[7]) / s
        qz = 0.25 * s
        qw = (rot[3] - rot[1]) / s
    return np.array((qw, qx, qy, qz))


def quaternion_conjugate(q):
    """
    https://stackoverflow.com/questions/15425313/inverse-quaternion

    :param q:
    :type q:
    :return:
    :rtype:
    """
    return np.array((-q[0], -q[1], -q[2]))


def quaternion_inverse(q):
    """
    Compute the inverse of a quaternion.

    The inverse of a quaternion is computed by taking the conjugate of the quaternion and dividing it by the norm of
    the quaternion.
    https://stackoverflow.com/questions/15425313/inverse-quaternion

    Parameters
    ----------
    q : np.ndarray
        Input quaternion.

    Returns
    -------
    np.ndarray
        Inverse of the input quaternion.
    """
    return quaternion_conjugate(q) / np.linalg.norm(q)


def quaternion_diff(q1, q2):
    """
    https://math.stackexchange.com/questions/2581668/
    error-measure-between-two-rotations-when-one-matrix-might-not-be-a-valid-rotatio

    Parameters
    ----------
    q1 : np.ndarray
        Quaternion 1.
    q2 : np.ndarray
        Quaternion 2.

    Returns
    -------
    float
        Difference between the two quaternions.
    """
    return np.linalg.norm(q1 * quaternion_inverse(q2) - 1)


def euler_angles_to_rotation_matrix(theta):
    """
    Determines the rotation matrix from the three Euler angles theta = [Psi, Theta, Phi] (in rad), which rotate the
    coordinate system in the order z, y', x''.

    Parameters
    ----------
    theta : np.ndarray
        (3) Euler angles in rad.

    Returns
    -------
    r : np.ndarray
        (3) Rotation matrix (z, y', x'').
    """

    # theta in rad
    r_x = np.array([[1., 0., 0.],
                    [0., math.cos(theta[0]), -math.sin(theta[0])],
                    [0., math.sin(theta[0]), math.cos(theta[0])]
                    ])

    r_y = np.array([[math.cos(theta[1]), 0, math.sin(theta[1])],
                    [0., 1., 0.],
                    [-math.sin(theta[1]), 0, math.cos(theta[1])]
                    ])

    r_z = np.array([[math.cos(theta[2]), -math.sin(theta[2]), 0],
                    [math.sin(theta[2]), math.cos(theta[2]), 0],
                    [0., 0., 1.]
                    ])

    r = np.dot(r_z, np.dot(r_y, r_x))

    return r


def rotation_matrix_to_euler_angles(r):
    """
    Calculates the euler angles theta = [Psi, Theta, Phi] (in rad) from the rotation matrix R which, rotate the
    coordinate system in the order z, y', x'' (https://www.learnopencv.com/rotation-matrix-to-euler-angles/).

    Parameters
    ----------
    r : np.ndarray
        (3, 3) Rotation matrix (z, y', x'').

    Returns
    -------
    theta : np.ndarray
        (3) Euler angles in rad.
    """
    sy = math.sqrt(r[0, 0] * r[0, 0] + r[1, 0] * r[1, 0])

    singular = sy < 1e-6

    if not singular:
        x = math.atan2(r[2, 1], r[2, 2])
        y = math.atan2(-r[2, 0], sy)
        z = math.atan2(r[1, 0], r[0, 0])
    else:
        x = math.atan2(-r[1, 2], r[1, 1])
        y = math.atan2(-r[2, 0], sy)
        z = 0

    return np.array([x, y, z])


def bases2rotmat(v1, v2):
    """
    Computes rotation matrix to rotate basis 1 (``v1``) to another basis (``v2``).

    Parameters
    ----------
    v1 : np.ndarray
        (3, 3) original basis.
    v2 : np.ndarray
        (3, 3) rotated basis.

    Returns
    -------
    rot_mat : np.ndarray
        (3, 3) rotation matrix to go from v1 to v2.
    """
    return np.linalg.solve(v1, v2).T


def rotate_matsimnibs_euler(axis, angle, matsimnibs, metric='rad'):
    """
    Rotates a matsimnibs matrix around ``axis`` by ``angle``.

    Parameters
    ----------
    axis : str
        One of ('x','y','z').
    angle : float
        Angle to rotate system around ``axis``.
    matsimnibs : np.ndarray
        (4, 4) SimNIBS matsimnibs coil orientation and position matrix.
    metric : str, default: 'rad'
        One of ('rad', 'deg'). If ``deg``, ``angle`` is transformed to radians.

    Returns
    -------
    rotated_matsimnibs : np.ndarray
        (4, 4) Rotated system.
    """

    if metric.lower().startswith('deg'):
        angle = np.deg2rad(angle)
    elif not metric.lower().startswith('rad'):
        raise Exception(ValueError)

    if axis == 'x':
        # rotate around x
        rotated_system = np.array((
            (1, 0, 0),
            (0, np.cos(angle), -np.sin(angle)),
            (0, np.sin(angle), np.cos(angle)),
        ))

    elif axis == 'y':
        # rotate around y
        rotated_system = np.array((
            (np.cos(angle), 0, np.sin(angle)),
            (0, 1, 0),
            (-np.sin(angle), 0, np.cos(angle)),
        ))
    elif axis == 'z':
        # rotate aroun z
        rotated_system = np.array((
            (np.cos(angle), -np.sin(angle), 0),
            (np.sin(angle), np.cos(angle), 0),
            (0, 0, 1),
        ))
    else:
        raise Exception(ValueError)

    rot_vecs = matsimnibs[0:3, :3].dot(rotated_system)

    return np.vstack((np.hstack((rot_vecs,
                                 matsimnibs[:3, 3, None])),
                      [0, 0, 0, 1]))


def rotmat_from_vecs(vec1, vec2):
    """
    Find the rotation matrix that aligns vec1 to vec2

    Parameters
    ----------
    vec 1: np.ndarray
        A 3D vector ('source').
    vec 2: np.ndarray
        A 3D vector ('destination').

    Returns
    -------
    rot_mat: scipy.Rotation
    A transform matrix (3x3) which when applied to vec1, aligns it with vec2.
    """
    a, b = (vec1 / np.linalg.norm(vec1)).reshape(3), (vec2 / np.linalg.norm(vec2)).reshape(3)
    v = np.cross(a, b)
    c = np.dot(a, b)
    s = np.linalg.norm(v)
    kmat = np.array([[0, -v[2], v[1]], [v[2], 0, -v[0]], [-v[1], v[0], 0]])
    rotation_matrix = np.eye(3) + kmat + kmat.dot(kmat) * ((1 - c) / (s ** 2))
    return Rotation.from_matrix(rotation_matrix)



""" Functions to interact with ANT / OMRON TMS Cobots """
import os
import h5py
import json
import datetime
import numpy as np
import pandas as pd

import pynibs


def merge_exp_data_cobot(subject, exp_idx, mesh_idx, coil_outlier_corr_cond=False,
                         remove_coil_skin_distance_outlier=True, coil_distance_corr=True,
                         verbose=False, plot=False):
    """
    Merge the TMS coil positions and the mep data into an experiment.hdf5 file.

    Parameters
    ----------
    subject : pynibs.subject.Subject
        Subject object.
    exp_idx : str
        Experiment ID.
    mesh_idx : str
        Mesh ID.
    coil_outlier_corr_cond : bool
        Correct outlier of coil position and orientation (+-2 mm, +-3 deg) in case of conditions.
    remove_coil_skin_distance_outlier : bool
        Remove outlier of coil position lying too far away from the skin surface (+- 5 mm).
    coil_distance_corr : bool
        Perform coil <-> head distance correction (coil is moved towards head surface until coil touches scalp).
    verbose : bool
        Plot output messages.
    plot : bool, optional, default: False
        Plot MEPs and p2p evaluation.
    """
    try:
        import biosig
    except ImportError:
        ImportError("Please install biosig from pynibs/pkg/biosig folder!")
        return

    fn_exp_hdf5 = subject.exp[exp_idx]["fn_exp_hdf5"][0]

    # 1) EMG
    cfs_fn = subject.exp[exp_idx]["fn_data"][0][0]
    cfs_header = json.loads(biosig.header(cfs_fn))
    cfs_emg = biosig.data(cfs_fn)

    num_sweeps = cfs_header["NumberOfSweeps"]

    total_num_samples = cfs_header["NumberOfSamples"]
    samples_per_sweep = int(total_num_samples / num_sweeps)
    sampling_rate = int(cfs_header["Samplingrate"])
    time = np.linspace(0, samples_per_sweep, samples_per_sweep) / sampling_rate
    num_channels = cfs_emg.shape[1]

    tms_pulse_time = subject.exp[exp_idx]['tms_pulse_time']
    # drop_mep_idx = None  # Which MEPs to remove before matching
    start_mep_ms = 18
    end_mep_ms = 40
    fn_plot = None

    # cfs_data_column = range(num_channels)
    d = dict()

    # get timestamps
    tms_pulse_timedelta = datetime.timedelta()
    # get hour, minute and second
    time_mep_list = []
    trigger_event_idcs = []
    # convert time string into integer
    for event in cfs_header["EVENT"]:
        date = datetime.datetime.strptime(event["TimeStamp"], '%Y-%b-%d %H:%M:%S')

        # we are interested in the tms pulse time, so add it to ts
        date += tms_pulse_timedelta
        time_mep_list.append(date)

        # compute indices in data block corresponding to the events
        if event["TYP"] == "0x7ffe":
            trigger_event_idcs.append(
                    round(event["POS"] * cfs_header["Samplingrate"])
            )

    num_sweeps = min(num_sweeps, len(trigger_event_idcs))

    for c_idx in range(num_channels):
        # Use emg data startinng from the index of the first trigger event
        # assumptions:
        # - after an initial offset all emg data were captured consecutively
        # - the first emg data frame may be captured without an explicit TMS
        #   tigger (eg. by checking the "write to disk" option)
        # - if we had dropouts in between the emg data block (not just at the
        #   beginning) we would need to adhere to the entire trigger_event_indices
        #   list.
        channel_emg = np.reshape(
                cfs_emg[trigger_event_idcs[0]:, c_idx],
                (num_sweeps, samples_per_sweep)
        )

        d[f"mep_raw_data_time_{c_idx}"] = []
        d[f"mep_filt_data_time_{c_idx}"] = []
        d[f"mep_raw_data_{c_idx}"] = []
        d[f"mep_filt_data_{c_idx}"] = []
        d[f"p2p_{c_idx}"] = []
        d[f"mep_latency_{c_idx}"] = []

        for s_idx in range(0, num_sweeps):
            if plot:
                fn_channel = os.path.join(os.path.split(subject.exp[exp_idx]["fn_data"][0][0])[0], "plots", str(c_idx))
                fn_plot = os.path.join(fn_channel, f"mep_{s_idx:04}")
                os.makedirs(fn_channel, exist_ok=True)

            # filter data and calculate p2p values
            p2p, mep_filt_data, latency = pynibs.calc_p2p(
                    sweep=channel_emg[s_idx],
                    tms_pulse_time=tms_pulse_time,
                    sampling_rate=sampling_rate,
                    start_mep=start_mep_ms,
                    end_mep=end_mep_ms,
                    measurement_start_time=0,
                    fn_plot=fn_plot
            )

            d[f"mep_raw_data_time_{c_idx}"].append(time)
            d[f"mep_filt_data_time_{c_idx}"].append(time)
            d[f"mep_raw_data_{c_idx}"].append(channel_emg[s_idx])
            d[f"mep_filt_data_{c_idx}"].append(mep_filt_data)
            d[f"p2p_{c_idx}"].append(p2p)
            d[f"mep_latency_{c_idx}"].append(latency)

    # 2) coil positions
    csv_fn = subject.exp[exp_idx]["fn_tms_nav"][0][0]
    out_coil_pos_fn = os.path.join(os.path.dirname(csv_fn), "coilpos")

    csv_coil_pos_data = pd.read_csv(csv_fn)
    centers = csv_coil_pos_data[['pos_x', 'pos_y', 'pos_z']].to_numpy()
    m0 = csv_coil_pos_data[['o_x', 'o_y', 'o_z']].to_numpy()
    m1 = csv_coil_pos_data[['n_x', 'n_y', 'n_z']].to_numpy()
    m2 = csv_coil_pos_data[['a_x', 'a_y', 'a_z']].to_numpy()

    matsimnibs = np.zeros((centers.shape[0], 4, 4))
    matsimnibs[:, 3, :3] = centers
    matsimnibs[:, 0, :3] = m0
    matsimnibs[:, 1, :3] = m1
    matsimnibs[:, 2, :3] = m2
    matsimnibs[:, 3, 3] = 1
    matsimnibs_t = matsimnibs.transpose()

    fn_matsimnibs_out = os.path.join(os.path.dirname(out_coil_pos_fn), "matsimnibs.hdf5")
    with h5py.File(fn_matsimnibs_out, 'w') as matsimnibs_hf5:
        matsimnibs_hf5["matsimnibs"] = matsimnibs_t

    '''
    # remove coil position outliers (in case of conditions)
    #######################################################
    if coil_outlier_corr_cond:
        if verbose:
            print("Removing coil position outliers")
        d = coil_outlier_correction_cond(exp=d,
                                         outlier_angle=5.,
                                         outlier_loc=3.,
                                         fn_exp_out=fn_exp_hdf5)
    '''

    # perform coil <-> head distance correction
    fn_mesh_hdf5 = subject.mesh[mesh_idx]['fn_mesh_hdf5']
    if coil_distance_corr:
        if verbose:
            print("Performing coil <-> head distance correction")
        matsimnibs_t = pynibs.coil_distance_correction_matsimnibs(matsimnibs=matsimnibs_t,
                                                           fn_mesh_hdf5=fn_mesh_hdf5,
                                                           distance=1,
                                                           remove_coil_skin_distance_outlier=remove_coil_skin_distance_outlier)

    d["coil_0"] = []
    for i in range(centers.shape[0]):
        d["coil_0"].append(matsimnibs_t[:, :, i])
    d["condition"] = csv_coil_pos_data['idx'].to_numpy()
    d["coil_mean"] = d["coil_0"]

    # 3) save in "experiment.hdf5"
    coil_0_np = np.array(d['coil_0'])
    pynibs.write_coil_pos_hdf5(
            fn_hdf=out_coil_pos_fn,
            centers=coil_0_np[:, :3, 3],  # np.array(centers),
            m0=coil_0_np[:, :3, 0],  # np.array(m0),
            m1=coil_0_np[:, :3, 1],  # np.array(m1),
            m2=coil_0_np[:, :3, 2],  # np.array(m2),
            overwrite=True
    )

    # create dictionary of stimulation data
    #######################################
    d_stim_data = dict()
    d_stim_data["coil_0"] = d["coil_0"]
    d_stim_data["number"] = csv_coil_pos_data["idx"].to_numpy()
    d_stim_data["current"] = csv_coil_pos_data["didt"].to_numpy()
    d_stim_data["date"] = csv_coil_pos_data["date"].to_list()
    d_stim_data["time"] = csv_coil_pos_data["time"].to_list()
    d_stim_data["coil_type"] = csv_coil_pos_data["coil_type"].to_list()

    # create dictionary of raw physiological data
    #############################################
    d_phys_data_raw = dict()
    # d_phys_data_raw["EMG Start"] = []
    # d_phys_data_raw["EMG End"] = []
    # d_phys_data_raw["EMG Res."] = []
    # d_phys_data_raw["EMG Channels"] = []
    d_phys_data_raw["EMG Window Start"] = [start_mep_ms] * num_sweeps
    d_phys_data_raw["EMG Window End"] = [end_mep_ms] * num_sweeps

    for chan in range(num_channels):
        d_phys_data_raw[f"mep_raw_data_time_{chan}"] = d[f"mep_raw_data_time_{chan}"]
        d_phys_data_raw[f"mep_raw_data_{chan}"] = d[f"mep_raw_data_{chan}"]

    # create dictionary of postprocessed physiological data
    #######################################################
    d_phys_data_postproc = dict()

    for chan in range(num_channels):
        d_phys_data_postproc[f"mep_filt_data_time_{chan}"] = d[f"mep_filt_data_time_{chan}"]
        d_phys_data_postproc[f"mep_filt_data_{chan}"] = d[f"mep_filt_data_{chan}"]
        d_phys_data_postproc[f"p2p_{chan}"] = d[f"p2p_{chan}"]
        d_phys_data_postproc[f"mep_latency_{chan}"] = d[f"mep_latency_{chan}"]

    # create pandas dataframes from dicts
    #####################################
    df_stim_data = pd.DataFrame.from_dict(d_stim_data)
    df_phys_data_raw = pd.DataFrame.from_dict(d_phys_data_raw)
    df_phys_data_postproc = pd.DataFrame.from_dict(d_phys_data_postproc)

    # save in .hdf5 file
    df_stim_data.to_hdf(fn_exp_hdf5, "stim_data")
    df_phys_data_raw.to_hdf(fn_exp_hdf5, "phys_data/raw/EMG")
    df_phys_data_postproc.to_hdf(fn_exp_hdf5, "phys_data/postproc/EMG")

    print("Done.")
